# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implements a high-level interface to Capella projects."""

from __future__ import annotations

import collections.abc as cabc
import enum
import functools
import sys
import typing as t
import warnings

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

VIRTUAL_NAMESPACE_PREFIX = (
    "https://dbinfrago.github.io/py-capellambse/virtual-namespace/"
)
"""The URI prefix to indicate a virtual namespace.

Classes defined in a virtual namespace can be used with :class:`Containment`,
:class:`Association`, :class:`Allocation` and :func:`wrap_xml` to provide
objects with special functionality.
"""

E = t.TypeVar("E", bound=enum.Enum)
"""TypeVar for ":py:class:`~enum.Enum`"."""
S = t.TypeVar("S", bound=str | None)
"""TypeVar for ":py:class:`str` | None"."""
T = t.TypeVar("T", bound="ModelObject")
"""TypeVar for ":py:class:`capellambse.model.ModelObject`"."""
T_co = t.TypeVar("T_co", bound="ModelObject", covariant=True)
"""Covariant TypeVar for ":py:class:`capellambse.model.ModelObject`"."""
U = t.TypeVar("U")
"""TypeVar (unbound)."""
U_co = t.TypeVar("U_co", covariant=True)
"""Covariant TypeVar (unbound)."""


@deprecated("set_accessor is deprecated and no longer needed")
def set_accessor(
    cls: type[ModelObject], attr: str, accessor: Accessor
) -> None:
    setattr(cls, attr, accessor)
    accessor.__set_name__(cls, attr)


@deprecated("set_self_references is deprecated, use a 'Containment' instead")
def set_self_references(*args: tuple[type[ModelObject], str]) -> None:
    for cls, attr in args:
        setattr(cls, attr, DirectProxyAccessor(cls, aslist=ElementList))  # type: ignore[deprecated]


@deprecated(
    '`@attr_equal("...")` is deprecated,'
    ' use `class X(ModelElement, eq="...")` instead'
)
def attr_equal(attr: str) -> cabc.Callable[[type[T]], type[T]]:
    def add_wrapped_eq(cls: type[T]) -> type[T]:
        orig_eq = cls.__eq__

        @functools.wraps(orig_eq)
        def new_eq(self, other: object) -> bool:
            try:
                cmpkey = getattr(self, attr)
            except AttributeError:
                pass
            else:
                if isinstance(other, type(cmpkey)):
                    result = cmpkey.__eq__(other)
                else:
                    result = other.__eq__(cmpkey)
                if result is not NotImplemented:
                    return result
            return orig_eq(self, other)

        cls.__eq__ = new_eq  # type: ignore[method-assign]
        cls.__hash__ = None  # type: ignore

        return cls

    return add_wrapped_eq


def stringy_enum(et: type[enum.Enum]) -> type[enum.Enum]:
    """Make an Enum stringy.

    This decorator makes an Enum's members compare equal to their
    respective ``name``.
    """

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self is other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __str__(self):
        return str(self.name)

    def __hash__(self):
        return hash(self.name)

    et.__eq__ = __eq__  # type: ignore[method-assign]
    et.__str__ = __str__  # type: ignore[method-assign]
    et.__hash__ = __hash__  # type: ignore[method-assign]
    return et


def reset_entrypoint_caches() -> None:
    """Reset all cached data from entrypoints."""
    for i in globals().values():
        if hasattr(i, "cache_clear"):
            i.cache_clear()


from . import diagram
from ._descriptors import *
from ._model import *
from ._obj import *
from ._pods import *

# NOTE: These are not in __all__ to avoid duplicate documentation in Sphinx,
#       however their docstring should mention the re-export.
from .diagram import AbstractDiagram as AbstractDiagram
from .diagram import Diagram as Diagram
from .diagram import DiagramAccessor as DiagramAccessor
from .diagram import DiagramType as DiagramType
from .diagram import DRepresentationDescriptor as DRepresentationDescriptor

if not t.TYPE_CHECKING:
    # Helper imports for Sphinx
    from lxml import etree  # noqa: F401

    from ._descriptors import __all__ as _all1
    from ._model import __all__ as _all2
    from ._obj import __all__ as _all3
    from ._pods import __all__ as _all4

    __all__ = [
        "E",
        "S",
        "T",
        "T_co",
        "U",
        "U_co",
        "VIRTUAL_NAMESPACE_PREFIX",
        "attr_equal",
        "diagram",
        "reset_entrypoint_caches",
        "set_self_references",
        "stringy_enum",
        *_all1,
        *_all2,
        *_all3,
        *_all4,
    ]

    _deprecated_names = {
        "GenericElement": ModelElement,
        "RoleTagAccessor": Containment,
        "AttrProxyAccessor": Association,
        "LinkAccessor": Allocation,
        "ReferenceSearchingAccessor": Backref,
    }

    def __getattr__(attr):
        if target := _deprecated_names.get(attr):
            warnings.warn(
                f"{attr} is deprecated, use {target.__name__} instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return target

        if attr == "XTYPE_ANCHORS":
            warnings.warn(
                f"{attr} is deprecated, use Namespace-based discovery instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return {}

        if attr == "XTYPE_HANDLERS":
            warnings.warn(
                f"{attr} is deprecated, use Namespace-based discovery instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return {
                None: [
                    cls[0][0]
                    for ns in enumerate_namespaces()
                    for cls in ns._classes.values()
                ]
            }

        raise AttributeError(f"{__name__} has no attribute {attr}")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implements the Capella Filtering extension."""

from __future__ import annotations

import operator
import subprocess
import sys
import typing as t
import warnings

import capellambse.model as m
from capellambse.metamodel import capellacore, capellamodeller

NS = m.Namespace(
    "http://www.polarsys.org/capella/filtering/{VERSION}",
    "filtering",
    "org.polarsys.capella.filtering",
    "7.0.0",
)

VIEWPOINT: t.Final = NS.viewpoint
NAMESPACE: t.Final = NS.uri.format(VERSION="6.0.0")
SYMBOLIC_NAME: t.Final = NS.alias


class FilteringModel(capellacore.NamedElement):
    criteria = m.Containment["FilteringCriterion"](
        "ownedFilteringCriteria", (NS, "FilteringCriterion")
    )
    criterion_pkgs = m.Containment["FilteringCriterionPkg"](
        "ownedFilteringCriterionPkgs", (NS, "FilteringCriterionPkg")
    )
    variability_features = m.Containment["FilteringCriterion"](
        "ownedVariabilityFeatures", (NS, "FilteringCriterion")
    )

    if not t.TYPE_CHECKING:
        criterion_packages = m.DeprecatedAccessor("criterion_pkgs")


class FilteringCriterion(capellacore.NamedElement):
    _xmltag = "ownedFilteringCriteria"

    filtered_objects = m.Backref[m.ModelElement](
        (), "filtering_criteria", legacy_by_type=True
    )


class FilteringCriterionSet(capellacore.NamedElement, abstract=True):
    criteria = m.Association["FilteringCriterion"](
        (NS, "FilteringCriterion"), "filteringCriteria"
    )
    variability_features = m.Association["FilteringCriterion"](
        (NS, "FilteringCriterion"), "variabilityFeatures"
    )


class FilteringResults(capellacore.NamedElement):
    results = m.Containment["AbstractFilteringResult"](
        "filteringResults", (NS, "AbstractFilteringResult")
    )
    result_pkgs = m.Containment["FilteringResultPkg"](
        "ownedFilteringResultPkgs", (NS, "FilteringResultPkg")
    )
    configurations = m.Containment["FilteringResult"](
        "configurations", (NS, "FilteringResult")
    )


class AbstractFilteringResult(capellacore.NamedElement, abstract=True):
    pass


class FilteringResult(FilteringCriterionSet, AbstractFilteringResult):
    """A filtering result."""


class AssociatedFilteringCriterionSet(FilteringCriterionSet):
    pass


class CreationDefaultFilteringCriterionSet(FilteringCriterionSet):
    pass


class FilteringResultPkg(capellacore.Namespace):
    results = m.Containment["AbstractFilteringResult"](
        "ownedFilteringResults", (NS, "AbstractFilteringResult")
    )
    packages = m.Containment["FilteringResultPkg"](
        "ownedFilteringResultPkgs", (NS, "FilteringResultPkg")
    )
    result_pkgs = m.Alias["m.ElementList[FilteringResultPkg]"]("packages")


class FilteringCriterionPkg(capellacore.Namespace):
    """A package containing multiple filtering criteria."""

    _xmltag = "ownedFilteringCriterionPkgs"

    criteria = m.Containment["FilteringCriterion"](
        "ownedFilteringCriteria", (NS, "FilteringCriterion")
    )
    packages = m.Containment["FilteringCriterionPkg"](
        "ownedFilteringCriterionPkgs", (NS, "FilteringCriterionPkg")
    )


class ComposedFilteringResult(AbstractFilteringResult):
    """A result obtained from boolean operations of other results."""

    union = m.Containment["UnionFilteringResultSet"](
        "UnionFilteringResultSet", (NS, "UnionFilteringResultSet")
    )
    intersection = m.Containment["IntersectionFilteringResultSet"](
        "IntersectionFilteringResultSet",
        (NS, "IntersectionFilteringResultSet"),
    )
    exclusion = m.Containment["ExclusionFilteringResultSet"](
        "ExclusionFilteringResultSet", (NS, "ExclusionFilteringResultSet")
    )


class FilteringResultSet(capellacore.NamedElement):
    results = m.Association["AbstractFilteringResult"](
        (NS, "AbstractFilteringResult"), "filteringResults"
    )


class UnionFilteringResultSet(FilteringResultSet):
    pass


class ExclusionFilteringResultSet(FilteringResultSet):
    pass


class IntersectionFilteringResultSet(FilteringResultSet):
    pass


def init() -> None:
    capellamodeller.SystemEngineering.filtering_model = m.Single(
        m.Filter("extensions", (NS, "FilteringModel"))
    )
    m.MelodyModel.filtering_model = property(  # type: ignore[attr-defined]
        operator.attrgetter("project.model_root.filtering_model")
    )
    m.ModelElement.filtering_criteria = property(
        lambda obj: obj.extensions.by_class(
            AssociatedFilteringCriterionSet
        ).map("criteria")
    )


def _main() -> None:
    if len(sys.argv) < 2:
        args = ["--help"]
    elif sys.argv[1] != "derive":
        print(f"Unsupported command: {sys.argv[1]}", file=sys.stderr)
        raise SystemExit(1)
    else:
        args = sys.argv[2:]

    warnings.warn(
        "This CLI entry point is deprecated, use 'capellambse filter-derive' instead",
        FutureWarning,
        stacklevel=2,
    )

    cmd = [sys.executable, "-mcapellambse", "filter-derive", *args]
    raise SystemExit(subprocess.run(cmd, check=False).returncode)


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum
import sys
import typing as t
import warnings

import capellambse.model as m

from . import activity, behavior, capellacore, information, modellingcore
from . import namespaces as ns

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

NS = ns.FA


# TODO Remove _AbstractExchange when removing deprecated features
class _AbstractExchange(m.ModelElement):
    if not t.TYPE_CHECKING:

        @property
        def source(self) -> m.ModelElement:
            raise TypeError(
                "AbstractExchange is deprecated and will be removed soon,"
                " use the concrete FunctionalExchange or ComponentExchange"
                " class or another common superclass instead"
            )

        @property
        def target(self) -> m.ModelElement:
            raise TypeError(
                "AbstractExchange is deprecated and will be removed soon,"
                " use the concrete FunctionalExchange or ComponentExchange"
                " class or another common superclass instead"
            )


@m.stringy_enum
@enum.unique
class ComponentExchangeKind(enum.Enum):
    """The kind of a ComponentExchange."""

    UNSET = "UNSET"
    """Communication kind is not set."""
    DELEGATION = "DELEGATION"
    """Indicates that the connector is a delegation connector."""
    ASSEMBLY = "ASSEMBLY"
    """Indicates that the connector is an assembly connector."""
    FLOW = "FLOW"
    """Describes a flow communication."""


@m.stringy_enum
@enum.unique
class ComponentPortKind(enum.Enum):
    STANDARD = "STANDARD"
    """Describes a standard port.

    A port is an interaction point between a Block or sub-Block and its
    environment that supports Exchanges with other ports.
    """
    FLOW = "FLOW"
    """Describes a flow port.

    A flow port is an interaction point through which input and/or
    output of items such as data, material, or energy may flow.
    """


@m.stringy_enum
@enum.unique
class ControlNodeKind(enum.Enum):
    OR = "OR"
    AND = "AND"
    ITERATE = "ITERATE"


@m.stringy_enum
@enum.unique
class FunctionKind(enum.Enum):
    """The kind of a Function."""

    FUNCTION = "FUNCTION"
    DUPLICATE = "DUPLICATE"
    GATHER = "GATHER"
    SELECT = "SELECT"
    SPLIT = "SPLIT"
    ROUTE = "ROUTE"


@m.stringy_enum
@enum.unique
class FunctionalChainKind(enum.Enum):
    """The kind of a Functional Chain."""

    SIMPLE = "SIMPLE"
    COMPOSITE = "COMPOSITE"
    FRAGMENT = "FRAGMENT"


@m.stringy_enum
@enum.unique
class OrientationPortKind(enum.Enum):
    """Direction of component ports."""

    UNSET = "UNSET"
    """The port orientation is undefined."""
    IN = "IN"
    """The port represents an input of the component it is used in."""
    OUT = "OUT"
    """The port represents an output of the component it is used in."""
    INOUT = "INOUT"
    """The port represents both an input and on output of the component."""


class AbstractFunctionalArchitecture(
    capellacore.ModellingArchitecture, abstract=True
):
    function_pkg = m.Single["FunctionPkg"](
        m.Containment("ownedFunctionPkg", (NS, "FunctionPkg"))
    )
    component_exchanges = m.Containment["ComponentExchange"](
        "ownedComponentExchanges", (NS, "ComponentExchange")
    )
    component_exchange_categories = m.Containment["ComponentExchangeCategory"](
        "ownedComponentExchangeCategories", (NS, "ComponentExchangeCategory")
    )
    functional_links = m.Containment["ExchangeLink"](
        "ownedFunctionalLinks", (NS, "ExchangeLink")
    )
    functional_allocations = m.Containment["ComponentFunctionalAllocation"](
        "ownedFunctionalAllocations", (NS, "ComponentFunctionalAllocation")
    )
    component_exchange_realizations = m.Containment[
        "ComponentExchangeRealization"
    ](
        "ownedComponentExchangeRealizations",
        (NS, "ComponentExchangeRealization"),
    )

    @property
    def all_functions(self) -> m.ElementList[AbstractFunction]:
        return self._model.search((NS, "AbstractFunction"), below=self)

    @property
    def all_functional_chains(self) -> m.ElementList[FunctionalChain]:
        return self._model.search((NS, "FunctionalChain"), below=self)

    @property
    def all_function_exchanges(self) -> m.ElementList[FunctionalExchange]:
        return self._model.search((NS, "FunctionalExchange"), below=self)

    if not t.TYPE_CHECKING:
        function_package = m.DeprecatedAccessor("function_pkg")


class AbstractFunctionalBlock(capellacore.ModellingBlock, abstract=True):
    functional_allocations = m.Containment["ComponentFunctionalAllocation"](
        "ownedFunctionalAllocation", (NS, "ComponentFunctionalAllocation")
    )
    allocated_functions = m.Allocation["AbstractFunction"](
        "ownedFunctionalAllocation",
        (NS, "ComponentFunctionalAllocation"),
        (NS, "AbstractFunction"),
        attr="targetElement",
        backattr="sourceElement",
    )
    component_exchanges = m.Containment["ComponentExchange"](
        "ownedComponentExchanges", (NS, "ComponentExchange")
    )
    component_exchange_categories = m.Containment["ComponentExchangeCategory"](
        "ownedComponentExchangeCategories", (NS, "ComponentExchangeCategory")
    )
    in_exchange_links = m.Association["ExchangeLink"](
        (NS, "ExchangeLink"), "inExchangeLinks"
    )
    out_exchange_links = m.Association["ExchangeLink"](
        (NS, "ExchangeLink"), "outExchangeLinks"
    )


class FunctionPkg(capellacore.Structure, abstract=True):
    functional_links = m.Containment["ExchangeLink"](
        "ownedFunctionalLinks", (NS, "ExchangeLink")
    )
    exchanges = m.Containment["FunctionalExchangeSpecification"](
        "ownedExchanges", (NS, "FunctionalExchangeSpecification")
    )
    exchange_specification_realizations = m.Containment[
        "ExchangeSpecificationRealization"
    ](
        "ownedExchangeSpecificationRealizations",
        (NS, "ExchangeSpecificationRealization"),
    )
    realized_exchange_specifications = m.Allocation["ExchangeSpecification"](
        "ownedExchangeSpecificationRealizations",
        (NS, "ExchangeSpecificationRealization"),
        (NS, "ExchangeSpecification"),
        attr="targetElement",
        backattr="sourceElement",
    )
    categories = m.Containment["ExchangeCategory"](
        "ownedCategories", (NS, "ExchangeCategory")
    )
    function_specifications = m.Containment["FunctionSpecification"](
        "ownedFunctionSpecifications", (NS, "FunctionSpecification")
    )


class FunctionSpecification(capellacore.Namespace, activity.AbstractActivity):
    in_exchange_links = m.Association["ExchangeLink"](
        (NS, "ExchangeLink"), "inExchangeLinks"
    )
    out_exchange_links = m.Association["ExchangeLink"](
        (NS, "ExchangeLink"), "outExchangeLinks"
    )
    ports = m.Containment["FunctionPort"](
        "ownedFunctionPorts", (NS, "FunctionPort")
    )


class ExchangeCategory(capellacore.NamedElement):
    _xmltag = "ownedCategories"

    exchanges = m.Association["FunctionalExchange"](
        (NS, "FunctionalExchange"), "exchanges"
    )


class ExchangeLink(capellacore.NamedRelationship):
    exchange_containment_links = m.Association["ExchangeContainment"](
        (NS, "ExchangeContainment"), "exchangeContainmentLinks"
    )
    exchange_containments = m.Containment["ExchangeContainment"](
        "ownedExchangeContainments", (NS, "ExchangeContainment")
    )
    sources = m.Association["FunctionSpecification"](
        (NS, "FunctionSpecification"), "sources"
    )
    destinations = m.Association["FunctionSpecification"](
        (NS, "FunctionSpecification"), "destinations"
    )


class ExchangeContainment(capellacore.Relationship):
    exchange = m.Single["ExchangeSpecification"](
        m.Association((NS, "ExchangeSpecification"), "exchange")
    )
    link = m.Single["ExchangeLink"](
        m.Association((NS, "ExchangeLink"), "link")
    )


class ExchangeSpecification(
    capellacore.NamedElement, activity.ActivityExchange, abstract=True
):
    link = m.Association["ExchangeContainment"](
        (NS, "ExchangeContainment"), "link"
    )


class FunctionalExchangeSpecification(ExchangeSpecification):
    pass


class FunctionalChain(
    capellacore.NamedElement,
    capellacore.InvolverElement,
    capellacore.InvolvedElement,
):
    _xmltag = "ownedFunctionalChains"

    kind = m.EnumPOD("kind", FunctionalChainKind)
    involvements = m.Containment["FunctionalChainInvolvement"](
        "ownedFunctionalChainInvolvements", (NS, "FunctionalChainInvolvement")
    )

    @property
    def involved_functions(self) -> m.ElementList[AbstractFunction]:
        objs = self.involvements.map("involved").by_class(AbstractFunction)
        return m.ElementList(self._model, objs._elements, legacy_by_type=True)

    @property
    def involved_links(self) -> m.ElementList[FunctionalExchange]:
        objs = self.involvements.map("involved").by_class(FunctionalExchange)
        return m.ElementList(self._model, objs._elements, legacy_by_type=True)

    @property
    def involved_chains(self) -> m.ElementList[FunctionalChain]:
        return self.involvements.map("involved").by_class(FunctionalChain)

    @property
    def involved(self) -> m.ElementList[AbstractFunction | FunctionalExchange]:
        objs = self.involvements.map("involved").by_class(
            AbstractFunction, FunctionalExchange
        )
        return m.ElementList(self._model, objs._elements, legacy_by_type=True)

    involving_chains = m.Backref["FunctionalChain"](
        (NS, "FunctionalChain"), "involved_chains"
    )

    functional_chain_realizations = m.Containment[
        "FunctionalChainRealization"
    ]("ownedFunctionalChainRealizations", (NS, "FunctionalChainRealization"))
    realized_chains = m.Allocation["FunctionalChain"](
        "ownedFunctionalChainRealizations",
        (NS, "FunctionalChainRealization"),
        (NS, "FunctionalChain"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_chains = m.Backref["FunctionalChain"](
        (NS, "FunctionalChain"), "realized_chains"
    )
    available_in_states = m.Association["capellacommon.State"](
        (ns.CAPELLACOMMON, "State"), "availableInStates"
    )
    precondition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "preCondition")
    )
    postcondition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "postCondition")
    )
    sequence_nodes = m.Containment["ControlNode"](
        "ownedSequenceNodes", (NS, "ControlNode")
    )
    sequence_links = m.Containment["SequenceLink"](
        "ownedSequenceLinks", (NS, "SequenceLink")
    )

    if not t.TYPE_CHECKING:
        control_nodes = m.DeprecatedAccessor("sequence_nodes")


class AbstractFunctionalChainContainer(
    capellacore.CapellaElement, abstract=True
):
    functional_chains = m.Containment["FunctionalChain"](
        "ownedFunctionalChains", (NS, "FunctionalChain")
    )


class FunctionalChainInvolvement(capellacore.Involvement, abstract=True):
    _xmltag = "ownedFunctionalChainInvolvements"


class FunctionalChainReference(FunctionalChainInvolvement):
    involved = m.Single["FunctionalChain"](
        m.Association((NS, "FunctionalChain"), None)
    )


class FunctionPort(
    information.Port,
    capellacore.TypedElement,
    behavior.AbstractEvent,
    abstract=True,
):
    represented_component_port = m.Single["ComponentPort"](
        m.Association((NS, "ComponentPort"), "representedComponentPort")
    )
    realized_ports = m.Allocation["FunctionPort"](
        None, None, (NS, "FunctionPort")
    )
    allocated_ports = m.Allocation["FunctionPort"](
        None, None, (NS, "FunctionPort")
    )
    exchanges = m.Backref["FunctionalExchange"](
        (NS, "FunctionalExchange"), "source", "target"
    )

    if not t.TYPE_CHECKING:
        owner = m.DeprecatedAccessor("parent")


class FunctionInputPort(FunctionPort, activity.InputPin):
    """A function input port."""

    _xmltag = "inputs"

    exchange_items = m.Association["information.ExchangeItem"](
        (ns.INFORMATION, "ExchangeItem"), "incomingExchangeItems"
    )


class FunctionOutputPort(FunctionPort, activity.OutputPin):
    """A function output port."""

    _xmltag = "outputs"

    exchange_items = m.Association["information.ExchangeItem"](
        (ns.INFORMATION, "ExchangeItem"), "outgoingExchangeItems"
    )


class AbstractFunctionAllocation(capellacore.Allocation, abstract=True):
    pass


class ComponentFunctionalAllocation(AbstractFunctionAllocation):
    pass


class FunctionalChainRealization(capellacore.Allocation):
    pass


class ExchangeSpecificationRealization(capellacore.Allocation, abstract=True):
    pass


class FunctionalExchangeRealization(capellacore.Allocation):
    pass


class FunctionRealization(AbstractFunctionAllocation):
    """A realization that links to a function."""

    _xmltag = "ownedFunctionRealizations"


class FunctionalExchange(
    capellacore.Relationship,
    capellacore.InvolvedElement,
    activity.ObjectFlow,
    behavior.AbstractEvent,
    information.AbstractEventOperation,
    # NOTE: NamedElement is first in the upstream metamodel,
    # but that would result in an MRO conflict with AbstractEventOperation,
    # which inherits from NamedElement.
    capellacore.NamedElement,
    _AbstractExchange,
):
    _xmltag = "ownedFunctionalExchanges"

    exchange_specifications = m.Association["FunctionalExchangeSpecification"](
        (NS, "FunctionalExchangeSpecification"), "exchangeSpecifications"
    )
    exchanged_items = m.Association["information.ExchangeItem"](
        (ns.INFORMATION, "ExchangeItem"), "exchangedItems"
    )
    functional_exchange_realizations = m.Containment[
        "FunctionalExchangeRealization"
    ](
        "ownedFunctionalExchangeRealizations",
        (NS, "FunctionalExchangeRealization"),
    )
    realized_functional_exchanges = m.Allocation["FunctionalExchange"](
        "ownedFunctionalExchangeRealizations",
        (NS, "FunctionalExchangeRealization"),
        (NS, "FunctionalExchange"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_functional_exchanges = m.Backref["FunctionalExchange"](
        (NS, "FunctionalExchange"), "realized_functional_exchanges"
    )
    owner = m.Single["ComponentExchange"](
        m.Backref((NS, "ComponentExchange"), "allocated_functional_exchanges")
    )
    allocating_component_exchange = m.Alias["ComponentExchange"]("owner")
    categories = m.Backref["ExchangeCategory"](
        (NS, "ExchangeCategory"), "exchanges"
    )

    involving_functional_chains = m.Backref["FunctionalChain"](
        (NS, "FunctionalChain"), "involved_links"
    )

    if not t.TYPE_CHECKING:
        exchange_items = m.DeprecatedAccessor("exchanged_items")


class AbstractFunction(
    capellacore.Namespace,
    capellacore.InvolvedElement,
    information.AbstractInstance,
    AbstractFunctionalChainContainer,
    activity.CallBehaviorAction,
    behavior.AbstractEvent,
    abstract=True,
):
    """An abstract function."""

    _xmltag = "ownedFunctions"

    kind = m.EnumPOD("kind", FunctionKind)
    condition = m.StringPOD("condition")
    functions = m.Containment["AbstractFunction"](
        "ownedFunctions", (NS, "AbstractFunction")
    )
    function_realizations = m.Containment["FunctionRealization"](
        "ownedFunctionRealizations", (NS, "FunctionRealization")
    )
    realized_functions = m.Allocation["AbstractFunction"](
        "ownedFunctionRealizations",
        (NS, "FunctionRealization"),
        (NS, "AbstractFunction"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_functions = m.Backref["AbstractFunction"](
        (NS, "AbstractFunction"), "realized_functions", legacy_by_type=True
    )
    exchanges = m.Containment["FunctionalExchange"](
        "ownedFunctionalExchanges", (NS, "FunctionalExchange")
    )
    available_in_states = m.Association["capellacommon.State"](
        (ns.CAPELLACOMMON, "State"), "availableInStates"
    )

    related_exchanges = m.Backref["FunctionalExchange"](
        (NS, "FunctionalExchange"), "source.owner", "target.owner"
    )
    scenarios = m.Backref["interaction.Scenario"](
        (ns.INTERACTION, "Scenario"), "related_functions"
    )

    @property
    def is_leaf(self) -> bool:
        return not self.functions


class ComponentExchange(
    behavior.AbstractEvent,
    information.AbstractEventOperation,
    # NOTE: NamedElement comes before ExchangeSpecification in the upstream
    # metamodel, but that would result in an MRO conflict.
    ExchangeSpecification,
    capellacore.NamedElement,
    _AbstractExchange,
):
    _xmltag = "ownedComponentExchanges"

    kind = m.EnumPOD("kind", ComponentExchangeKind)
    is_oriented = m.BoolPOD("oriented")

    functional_exchange_allocations = m.Containment[
        "ComponentExchangeFunctionalExchangeAllocation"
    ](
        "ownedComponentExchangeFunctionalExchangeAllocations",
        (NS, "ComponentExchangeFunctionalExchangeAllocation"),
    )
    allocated_functional_exchanges = m.Allocation["FunctionalExchange"](
        "ownedComponentExchangeFunctionalExchangeAllocations",
        (NS, "ComponentExchangeFunctionalExchangeAllocation"),
        (NS, "FunctionalExchange"),
        attr="targetElement",
        backattr="sourceElement",
    )
    component_exchange_realizations = m.Containment[
        "ComponentExchangeRealization"
    ](
        "ownedComponentExchangeRealizations",
        (NS, "ComponentExchangeRealization"),
    )
    realized_component_exchanges = m.Allocation["ComponentExchange"](
        "ownedComponentExchangeRealizations",
        (NS, "ComponentExchangeRealization"),
        (NS, "ComponentExchange"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_component_exchanges = m.Backref["ComponentExchange"](
        (NS, "ComponentExchange"), "realized_component_exchanges"
    )
    ends = m.Containment["ComponentExchangeEnd"](
        "ownedComponentExchangeEnds", (NS, "ComponentExchangeEnd")
    )
    categories = m.Backref["ComponentExchangeCategory"](
        (NS, "ComponentExchangeCategory"), "exchanges"
    )

    allocating_physical_links = m.Backref["cs.PhysicalLink"](
        (ns.CS, "PhysicalLink"), "allocated_component_exchanges"
    )
    allocating_physical_paths = m.Backref["cs.PhysicalPath"](
        (ns.CS, "PhysicalPath"), "allocated_component_exchanges"
    )

    @property
    @deprecated(
        (
            "ComponentExchange.allocating_physical_link is deprecated,"
            " because it only takes into account the first allocation link."
            " Use allocating_physical_links instead, which allows multiple links."
        ),
        category=FutureWarning,
    )
    def allocating_physical_link(self) -> cs.PhysicalLink | None:
        links = self.allocating_physical_links
        return links[0] if links else None

    @property
    @deprecated(
        (
            "ComponentExchange.owner is deprecated,"
            " because it only takes into account the first allocation link."
            " Use allocating_physical_links instead, which allows multiple links."
        ),
        category=FutureWarning,
    )
    def owner(self) -> cs.PhysicalLink | None:
        links = self.allocating_physical_links
        return links[0] if links else None

    @property
    def exchange_items(
        self,
    ) -> m.ElementList[modellingcore.AbstractExchangeItem]:
        return (
            self.convoyed_informations
            + self.allocated_functional_exchanges.map("exchanged_items")
        )

    if not t.TYPE_CHECKING:
        allocated_exchange_items = m.DeprecatedAccessor(
            "convoyed_informations"
        )


class ComponentExchangeAllocation(capellacore.Allocation):
    pass


class ComponentExchangeAllocator(capellacore.NamedElement, abstract=True):
    component_exchange_allocations = m.Containment[
        "ComponentExchangeAllocation"
    ]("ownedComponentExchangeAllocations", (NS, "ComponentExchangeAllocation"))
    allocated_component_exchanges = m.Allocation["ComponentExchange"](
        "ownedComponentExchangeAllocations",
        (NS, "ComponentExchangeAllocation"),
        (NS, "ComponentExchange"),
        attr="targetElement",
        backattr="sourceElement",
    )


class ComponentExchangeCategory(capellacore.NamedElement):
    _xmltag = "ownedComponentExchangeCategories"

    exchanges = m.Association["ComponentExchange"](
        (NS, "ComponentExchange"), "exchanges"
    )


class ComponentExchangeEnd(
    modellingcore.InformationsExchanger, capellacore.CapellaElement
):
    port = m.Single["information.Port"](
        m.Association((ns.INFORMATION, "Port"), "port")
    )
    part = m.Single["cs.Part"](m.Association((ns.CS, "Part"), "part"))


class ComponentExchangeFunctionalExchangeAllocation(
    AbstractFunctionAllocation
):
    pass


class ComponentExchangeRealization(ExchangeSpecificationRealization):
    pass


class ComponentPort(
    information.Port, modellingcore.InformationsExchanger, information.Property
):
    """A component port."""

    _xmltag = "ownedFeatures"

    orientation = m.EnumPOD("orientation", OrientationPortKind)
    kind = m.EnumPOD("kind", ComponentPortKind)
    exchanges = m.Backref["ComponentExchange"](
        (NS, "ComponentExchange"), "source", "target"
    )

    if not t.TYPE_CHECKING:
        direction = m.DeprecatedAccessor("orientation")
        owner = m.DeprecatedAccessor("parent")


class ComponentPortAllocation(capellacore.Allocation):
    ends = m.Containment["ComponentPortAllocationEnd"](
        "ownedComponentPortAllocationEnds", (NS, "ComponentPortAllocationEnd")
    )


class ComponentPortAllocationEnd(capellacore.CapellaElement):
    port = m.Single["information.Port"](
        m.Association((ns.INFORMATION, "Port"), "port")
    )
    part = m.Single["cs.Part"](m.Association((ns.CS, "Part"), "part"))


class ReferenceHierarchyContext(modellingcore.ModelElement, abstract=True):
    source_reference_hierarchy = m.Association["FunctionalChainReference"](
        (NS, "FunctionalChainReference"), "sourceReferenceHierarchy"
    )
    target_reference_hierarchy = m.Association["FunctionalChainReference"](
        (NS, "FunctionalChainReference"), "targetReferenceHierarchy"
    )


class FunctionalChainInvolvementLink(
    FunctionalChainInvolvement, ReferenceHierarchyContext
):
    exchange_context = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "exchangeContext")
    )
    exchanged_items = m.Association["information.ExchangeItem"](
        (ns.INFORMATION, "ExchangeItem"), "exchangedItems"
    )
    source = m.Single["FunctionalChainInvolvementFunction"](
        m.Association((NS, "FunctionalChainInvolvementFunction"), "source")
    )
    target = m.Single["FunctionalChainInvolvementFunction"](
        m.Association((NS, "FunctionalChainInvolvementFunction"), "target")
    )

    if not t.TYPE_CHECKING:
        context = m.DeprecatedAccessor("exchange_context")


class SequenceLink(capellacore.CapellaElement, ReferenceHierarchyContext):
    condition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "condition")
    )
    links = m.Association["FunctionalChainInvolvementLink"](
        (NS, "FunctionalChainInvolvementLink"), "links"
    )
    source = m.Single["SequenceLinkEnd"](
        m.Association((NS, "SequenceLinkEnd"), "source")
    )
    target = m.Single["SequenceLinkEnd"](
        m.Association((NS, "SequenceLinkEnd"), "target")
    )


class SequenceLinkEnd(capellacore.CapellaElement, abstract=True):
    pass


class FunctionalChainInvolvementFunction(
    FunctionalChainInvolvement, SequenceLinkEnd
):
    pass


class ControlNode(SequenceLinkEnd):
    _xmltag = "ownedSequenceNodes"

    kind = m.EnumPOD("kind", ControlNodeKind)


if not t.TYPE_CHECKING:

    def __getattr__(attr):
        if attr == "AbstractExchange":
            warnings.warn(
                (
                    "AbstractExchange is deprecated and will be removed soon,"
                    " use the concrete FunctionalExchange or ComponentExchange"
                    " class or another common superclass instead"
                ),
                DeprecationWarning,
                stacklevel=2,
            )
            return _AbstractExchange

        if attr == "Function":
            warnings.warn(
                "Function has been merged into AbstractFunction",
                DeprecationWarning,
                stacklevel=2,
            )
            return AbstractFunction

        raise AttributeError(f"{__name__} has no attribute {attr}")


from . import capellacommon, cs, interaction  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections.abc as cabc
import errno
import io
import logging
import math
import os
import pathlib
import sys
import typing as t
import urllib.parse
import weakref

import diskcache
import requests
import requests.exceptions
import urllib3.exceptions

import capellambse
from capellambse import helpers

from . import abc

LOGGER = logging.getLogger(__name__)
MAX_SEARCHED_JOBS = (
    int(os.environ.get("CAPELLAMBSE_GLART_MAX_JOBS", "1000")) or sys.maxsize
)


class GitlabArtifactsFiles(abc.FileHandler):
    """Download files from Gitlab's artifacts hosting service.

    This file handler is roughly equivalent to an HTTPFileHandler with
    appropriate headers and the following URL::

        https://<path>/api/v4/projects/<project>/jobs/artifacts/<branch>/raw/<subdir>/%s?job_name=<job>

    One important difference is that this file handler will always use
    the latest successful job, regardless of the overall pipeline
    status, while an HTTPFileHandler with the above URL would only
    consider jobs from successful pipelines.

    This file handler uses several of the Gitlab CI/CD `pre-defined
    environment variables`__. Refer to the documentation for their exact
    meaning and behavior during CI runs, and see below for how they are
    used.

    __ https://docs.gitlab.com/ee/ci/variables/predefined_variables.html

    Parameters
    ----------
    path
        The URL to fetch artifacts from, in one of the following formats:

        -   The literal string ``glart:`` (or ``glart://``), which uses
            the URL from ``$CI_SERVER_FQDN`` or - if that is not set -
            the public Gitlab instance at ``https://gitlab.com``.
        -   The base URL of the Gitlab server, using ``glart:``,
            ``glart+http:`` or ``glart+https:`` as the protocol.
            ``glart:`` uses the protocol specified by
            ``$CI_SERVER_PROTOCOL``, or falls back to HTTPS.
        -   A URL that combines the above with some of the required
            parameters described below, using the format::

                glart://gitlab.mycompany.com/group/subgroup/project#branch=<branch>&job=<jobname>

            Note that this format does not support numeric IDs for the
            project and job, thus requiring to pass a branch name. Any
            of the parts of this combined URL may instead be passed
            explicitly via keyword arguments.

            Keyword arguments have precedence over the combined URL.
    token
        A personal or project access token with ``read_api`` permission
        on the specified project. The following ways are supported for
        passing the token, which are checked in order:

        1. Directly via this argument.
        2. If the argument starts with a dollar sign (``$``), it is
           treated as the name of an environment variable that points to
           a file containing the token. This is compatible with
           variables of type "File" in Gitlab CI/CD.
        3. A file called ``gitlab_artifacts_token`` in the
           ``$CREDENTIALS_DIRECTORY``.
        4. The ``CI_JOB_TOKEN`` environment variable. This is intended
           for use in Gitlab pipelines, in order to avoid having to
           create explicit tokens. Note that your instance might be set
           up with restrictive default permissions for the job token.
    project
        The path (e.g. ``group/subgroup/project``) or numeric ID of the
        project to pull the artifacts from. Defaults to the
        ``$CI_PROJECT_ID`` environment variable, which Gitlab sets to
        the project currently executing a pipeline.
    branch
        The branch to pull artifacts from. Defaults to the value of the
        ``CI_DEFAULT_BRANCH`` environment variable, or ``main`` if that
        is unset. Ignored if a numeric ID is given for ``job``.
    job
        Name of the job to pull artifacts from. May also be a numeric
        job ID. Defaults to "update_capella_diagram_cache".

        If a job name was given, the Gitlab API is queried for the most
        recent successful job on the given ``branch`` that has attached
        artifacts. Note that jobs whose artifacts have been deleted (for
        example, because their retention period expired) are ignored.

        By default, at most 1000 jobs will be checked. This includes all
        successful jobs in the repo, regardless of their name or the
        branch they ran on. This number can be changed using the
        ``CAPELLAMBSE_GLART_MAX_JOBS`` environment variable.
    subdir
        An optional path prefix inside the artifacts archive to prepend
        to all file names.
    disable_cache
        Clear the local cache and discard any previously cached data.

    See Also
    --------
    capellambse.filehandler.http.HTTPFileHandler :
        A general-purpose HTTP file handler.
    """

    def __init__(
        self,
        path: str,
        *,
        subdir: str | pathlib.PurePosixPath = "/",
        token: str | None = None,
        project: str | int | None = None,
        branch: str | None = None,
        job: str | int | None = None,
        disable_cache: bool = False,
    ) -> None:
        path = str(path)
        super().__init__(path, subdir=subdir)

        self.__session = requests.Session()

        self.__path, urlparts = self.__resolve_path(path)
        if not project:
            project = urlparts["project"]
        if not branch:
            branch = urlparts["branch"]
        if not job:
            job = urlparts["job"]
        if str(self.subdir) in ("", ".", "/"):
            self.subdir = helpers.normalize_pure_path(urlparts["subdir"])

        self.__token = self.__resolve_token(token)
        self.__project = self.__resolve_project(project)
        self.__branch = branch or os.getenv("CI_DEFAULT_BRANCH") or "main"
        self.__job, job = self.__resolve_job(job)

        self.path = (
            f"glart+{self.__path}/{project}/-/{self.subdir}"
            f"#branch={branch}&job={job}"
        )
        self.__cache = diskcache.Cache(
            capellambse.dirs.user_cache_path / "gitlab-artifacts"
        )
        self.__fnz = weakref.finalize(self, self.__cache.close)

        if disable_cache:
            self.__cache.clear()

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(path={self.__path!r}, token=<HIDDEN>,"
            f" project={self.__project!r}, branch={self.__branch!r},"
            f" job={self.__job!r}, subdir={self.subdir!r})"
        )

    @staticmethod
    def __resolve_path(path: str) -> tuple[str, _URLParts]:
        scheme, netloc, project, _, _, fragment = urllib.parse.urlparse(path)
        scheme = scheme.removeprefix("glart+")

        if netloc:
            if scheme == "glart":
                scheme = os.getenv("CI_SERVER_PROTOCOL", "https")
            path = urllib.parse.urlunparse((scheme, netloc, "", "", "", ""))
        elif path := os.getenv("CI_SERVER_URL", ""):
            LOGGER.debug("Using current Gitlab instance: %s", path)
        else:
            path = "https://gitlab.com"
            LOGGER.debug("Using public Gitlab instance at: %s", path)

        if "/-/" in project:
            project, subdir = project.split("/-/", 1)
        else:
            subdir = ""
        args = dict(i.split("=", 1) for i in fragment.split("&") if "=" in i)
        urlparts = _URLParts(
            project=project.strip("/").removesuffix(".git"),
            branch=args.get("branch", ""),
            job=args.get("job", ""),
            subdir=subdir.strip("/"),
        )
        return path, urlparts

    @classmethod
    def __resolve_token(cls, token: str | None) -> str:
        if token:
            return cls.__load_token_from_arg(token)

        if cred_dir := os.getenv("CREDENTIALS_DIRECTORY"):
            token = cls.__load_token_from_credentials(cred_dir)
            if token:
                return token

        if token := os.getenv("CI_JOB_TOKEN"):
            LOGGER.debug("Using the $CI_JOB_TOKEN")
            return token

        raise TypeError("Cannot connect to Gitlab: No 'token' found")

    @staticmethod
    def __load_token_from_arg(token: str) -> str:
        if not token.startswith("$"):
            return token

        try:
            filename = os.environ[token[1:]]
        except KeyError:
            raise ValueError(
                f"Token environment variable {token} is unset"
            ) from None

        try:
            contents = pathlib.Path(filename).read_text(encoding="ascii")
        except FileNotFoundError:
            raise ValueError(
                f"Token file from {token} not found"
                " - did you set the variable type to FILE?"
            ) from None
        except OSError:
            raise ValueError(f"Cannot read token file from {token}") from None

        contents = contents.strip()
        if not contents:
            raise ValueError(f"Token file from {token} is empty")
        return contents

    @staticmethod
    def __load_token_from_credentials(cred_dir: str) -> str | None:
        cred_file = pathlib.Path(cred_dir, "gitlab_artifacts_token")
        try:
            token = cred_file.read_text(encoding="locale").strip()
        except OSError as err:
            LOGGER.debug("Cannot read token from %r: %s", cred_file, err)
        else:
            if token:
                LOGGER.debug("Using token from %s", cred_file)
                return token
            LOGGER.debug("Token file is empty: %s", cred_file)
        return None

    def __resolve_project(self, project: str | int | None) -> int:
        if isinstance(project, int):
            return project

        if project:
            project = urllib.parse.quote(project, safe="")
            info = self.__get(f"/projects/{project}")
            LOGGER.debug("Resolved project %r to ID %d", project, info["id"])
            return info["id"]

        if project := os.getenv("CI_PROJECT_ID"):
            return int(project)

        raise TypeError("No 'project' specified, and not running in Gitlab CI")

    def __resolve_job(self, job: str | int) -> tuple[int, str]:
        try:
            job = int(job)
            return job, str(job)
        except ValueError:
            pass
        assert isinstance(job, str)

        if not job:
            job = "update_capella_diagram_cache"

        for jobinfo in self.__iterget(
            f"/projects/{self.__project}/jobs?scope=success",
            max=MAX_SEARCHED_JOBS,
        ):
            if (
                jobinfo["name"] == job
                and jobinfo["pipeline"]["ref"] == self.__branch
                and any(
                    i["file_type"] == "archive"
                    for i in jobinfo.get("artifacts", [])
                )
            ):
                LOGGER.debug("Selected job with ID %d", jobinfo["id"])
                return (jobinfo["id"], job)

        raise RuntimeError(
            f"No recent successful {job!r} job found on {self.__branch!r}"
        )

    def __rawget(self, url: str) -> requests.Response:
        """Make a GET request and return the raw Response object.

        Parameters
        ----------
        url
            The full URL to request, including the server and protocol.
        """
        LOGGER.debug("GET %s", url)
        return self.__session.get(
            url, headers={"PRIVATE-TOKEN": self.__token}, timeout=30
        )

    def __get(self, path: str) -> t.Any:
        """Make a GET request and return the decoded JSON response."""
        response = self.__rawget(f"{self.__path}/api/v4{path}")
        LOGGER.debug("Status: %d", response.status_code)
        response.raise_for_status()
        return response.json()

    def __iterget(self, path: str, *, max: int = 0) -> cabc.Iterator[t.Any]:
        """Iterate over a result set, using multiple GET requests if necessary.

        This method implements the keyset-based pagination method.

        Parameters
        ----------
        path
            The base URL for the first request.
        max
            Maximum number of entries to yield. If 0, yield everything.
        """
        sep = ("?", "&")["?" in path]
        next_url = (
            f"{self.__path}/api/v4{path}{sep}pagination=keyset&per_page="
            + str(min(100, max or 100))
        )
        stop = max or math.inf

        while next_url:
            response = self.__rawget(next_url)
            response.raise_for_status()
            for i, object in enumerate(response.json()):
                yield object
                if i >= stop:
                    return

            next_url = response.links.get("next", {}).get("url")

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        path = str(helpers.normalize_pure_path(filename, base=self.subdir))

        if "w" in mode:
            raise TypeError("Cannot write to Gitlab artifacts")

        cachekey = f"{self.__path}|{self.__project}|{self.__job}|{path}"
        try:
            content = self.__cache[cachekey]
        except KeyError:
            pass
        else:
            if content is None:
                LOGGER.debug("Negative cache hit for %r", path)
                raise FileNotFoundError(errno.ENOENT, filename)
            LOGGER.debug("Opening cached file %r for reading", path)
            return io.BytesIO(content)

        try:
            response = self.__rawget(
                f"{self.__path}/api/v4/projects/{self.__project}"
                f"/jobs/{self.__job}/artifacts/{path}"
            )
        # <https://gitlab.com/gitlab-org/gitlab/-/issues/414807>
        except requests.exceptions.ChunkedEncodingError as err:
            if len(err.args) != 1:
                raise
            (err1,) = err.args
            if not isinstance(err1, urllib3.exceptions.ProtocolError):
                raise
            if len(err1.args) != 2:
                raise
            (_, err2) = err1.args
            if not isinstance(err2, urllib3.exceptions.IncompleteRead):
                raise
            if err2.args != (0, 2):
                raise

            LOGGER.debug("File not found in artifacts archive: %r", path)
            self.__cache[cachekey] = None
            raise FileNotFoundError(errno.ENOENT, filename) from None
        if response.status_code in (400, 404):
            LOGGER.debug("File not found in artifacts archive: %r", path)
            self.__cache[cachekey] = None
            raise FileNotFoundError(errno.ENOENT, filename)
        response.raise_for_status()
        LOGGER.debug("Opening file %r for reading", path)
        self.__cache[cachekey] = response.content
        return io.BytesIO(response.content)


class _URLParts(t.TypedDict):
    project: str
    branch: str
    job: str
    subdir: str



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum
import typing as t

import capellambse.model as m

from .. import behavior, capellacore
from .. import namespaces as ns
from . import datavalue

NS = ns.INFORMATION_COMMUNICATION


@m.stringy_enum
@enum.unique
class CommunicationLinkKind(enum.Enum):
    """Enumeration listing the various possibilities of communication links."""

    UNSET = "UNSET"
    """The CommunicationLink protocol is not yet set."""
    PRODUCE = "PRODUCE"
    """The CommunicationLink describes a production of ExchangeItem."""
    CONSUME = "CONSUME"
    """The CommunicationLink describes a comsumption of ExchangeItem."""
    SEND = "SEND"
    """The CommunicationLink describes a sending of ExchangeItem."""
    RECEIVE = "RECEIVE"
    """The CommunicationLink describes a reception of ExchangeItem."""
    CALL = "CALL"
    """The CommunicationLink describes a call of ExchangeItem."""
    EXECUTE = "EXECUTE"
    """The CommunicationLink describes an execution of ExchangeItem."""
    WRITE = "WRITE"
    """The CommunicationLink describes a writing of ExchangeItem."""
    ACCESS = "ACCESS"
    """The CommunicationLink describes an access to the ExchangeItem."""
    ACQUIRE = "ACQUIRE"
    """The CommunicationLink describes an acquisition of ExchangeItem."""
    TRANSMIT = "TRANSMIT"
    """The CommunicationLink describes a transmission of ExchangeItem."""


@m.stringy_enum
@enum.unique
class CommunicationLinkProtocol(enum.Enum):
    """The various possibilities for the protocol of the communication link."""

    UNSET = "UNSET"
    """The CommunicationLink protocol is not yet set."""
    UNICAST = "UNICAST"
    """Describes sending an ExchangeItem using the unicast protocol."""
    MULTICAST = "MULTICAST"
    """Describes sending an ExchangeItem using the multicast protocol."""
    BROADCAST = "BROADCAST"
    """Describes sending an ExchangeItem using the broadcast protocol."""
    SYNCHRONOUS = "SYNCHRONOUS"
    """Describes a call of the ExchangeItem using the synchronous protocol."""
    ASYNCHRONOUS = "ASYNCHRONOUS"
    """Describes a call of the ExchangeItem using the asynchronous protocol."""
    READ = "READ"
    """Describes access to the ExchangeItem by reading it."""
    ACCEPT = "ACCEPT"
    """Describes access to the ExchangeItem by accepting it."""


class CommunicationItem(
    capellacore.Classifier, datavalue.DataValueContainer, abstract=True
):
    visibility = m.EnumPOD("visibility", capellacore.VisibilityKind)
    state_machines = m.Containment["capellacommon.StateMachine"](
        "ownedStateMachines", (ns.CAPELLACOMMON, "StateMachine")
    )


class Exception(CommunicationItem):
    pass


class Message(CommunicationItem):
    pass


class MessageReference(capellacore.Relationship):
    message = m.Single["Message"](m.Association((NS, "Message"), "message"))


class MessageReferencePkg(capellacore.Structure, abstract=True):
    message_references = m.Containment["MessageReference"](
        "ownedMessageReferences", (NS, "MessageReference")
    )


class Signal(CommunicationItem, behavior.AbstractSignal):
    instances = m.Containment["SignalInstance"](
        "signalInstances", (NS, "SignalInstance")
    )


from . import AbstractInstance


class SignalInstance(AbstractInstance):
    pass


class CommunicationLink(capellacore.CapellaElement):
    kind = m.EnumPOD("kind", CommunicationLinkKind)
    protocol = m.EnumPOD("protocol", CommunicationLinkProtocol)
    exchange_item = m.Association["ExchangeItem"](
        (NS, "ExchangeItem"), "exchangeItem"
    )


class CommunicationLinkExchanger(m.ModelElement, abstract=True):
    links = m.Containment["CommunicationLink"](
        "ownedCommunicationLinks", (NS, "CommunicationLink")
    )


if t.TYPE_CHECKING:
    from . import ExchangeItem  # noqa: F401
from .. import capellacommon  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum
import typing as t
import warnings

import capellambse.model as m

from . import behavior, capellacore, fa
from . import namespaces as ns

NS = ns.INTERACTION


@m.stringy_enum
@enum.unique
class InteractionOperatorKind(enum.Enum):
    ALT = "ALT"
    OPT = "OPT"
    PAR = "PAR"
    LOOP = "LOOP"
    CRITICAL = "CRITICAL"
    NEG = "NEG"
    ASSERT = "ASSERT"
    STRICT = "STRICT"
    SEQ = "SEQ"
    IGNORE = "IGNORE"
    CONSIDER = "CONSIDER"
    UNSET = "UNSET"


@m.stringy_enum
@enum.unique
class MessageKind(enum.Enum):
    """Identifies the type of message.

    This concept is similar to UML ``MessageSort``.
    """

    UNSET = "UNSET"
    """The message kind is not specified."""
    ASYNCHRONOUS_CALL = "ASYNCHRONOUS_CALL"
    """The message was generated by an asynchronous call to an operation.

    Equivalent to UML ``MessageSort::asynchCall``.
    """
    SYNCHRONOUS_CALL = "SYNCHRONOUS_CALL"
    """The message was generated by a synchronous call to an operation.

    Equivalent to UML ``MessageSort::synchCall``.
    """
    REPLY = "REPLY"
    """The message is a reply message to an operation call.

    Equivalent to UML ``MessageSort::reply``.
    """
    DELETE = "DELETE"
    """The message designates the termination of another lifeline.

    Equivalent to UML ``MessageSort::deleteMessage``.
    """
    CREATE = "CREATE"
    """The message designates the creation of an instance role."""
    TIMER = "TIMER"


@m.stringy_enum
@enum.unique
class ScenarioKind(enum.Enum):
    INTERFACE = "INTERFACE"
    DATA_FLOW = "DATA_FLOW"
    INTERACTION = "INTERACTION"
    FUNCTIONAL = "FUNCTIONAL"
    UNSET = "UNSET"


class SequenceMessage(capellacore.NamedElement):
    """A sequence message."""

    kind = m.EnumPOD("kind", MessageKind)
    exchange_context = m.Association["capellacore.Constraint"](
        (ns.CAPELLACORE, "Constraint"), "exchangeContext"
    )
    sending_end = m.Association["MessageEnd"]((NS, "MessageEnd"), "sendingEnd")
    receiving_end = m.Association["MessageEnd"](
        (NS, "MessageEnd"), "receivingEnd"
    )
    exchanged_items = m.Association["information.ExchangeItem"](
        (ns.INFORMATION, "ExchangeItem"), "exchangedItems"
    )
    valuations = m.Containment["SequenceMessageValuation"](
        "ownedSequenceMessageValuations", (NS, "SequenceMessageValuation")
    )

    if not t.TYPE_CHECKING:
        source = m.DeprecatedAccessor("sending_end")
        target = m.DeprecatedAccessor("receiving_end")


class Scenario(capellacore.Namespace, behavior.AbstractBehavior):
    kind = m.EnumPOD("kind", ScenarioKind)
    is_merged = m.BoolPOD("merged")

    precondition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "preCondition")
    )
    postcondition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "postCondition")
    )
    instance_roles = m.Containment["InstanceRole"](
        "ownedInstanceRoles", (NS, "InstanceRole")
    )
    messages = m.Containment["SequenceMessage"](
        "ownedMessages", (NS, "SequenceMessage")
    )
    fragments = m.Containment["InteractionFragment"](
        "ownedInteractionFragments",
        (NS, "InteractionFragment"),
        legacy_by_type=True,
    )
    time_lapses = m.Containment["TimeLapse"](
        "ownedTimeLapses", (NS, "TimeLapse"), legacy_by_type=True
    )
    events = m.Containment["Event"](
        "ownedEvents", (NS, "Event"), legacy_by_type=True
    )
    formal_gates = m.Containment["Gate"]("ownedFormalGates", (NS, "Gate"))
    scenario_realizations = m.Containment["ScenarioRealization"](
        "ownedScenarioRealization", (NS, "ScenarioRealization")
    )
    realized_scenarios = m.Allocation["Scenario"](
        "ownedScenarioRealization",
        (NS, "ScenarioRealization"),
        (NS, "Scenario"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_scenarios = m.Backref["Scenario"](
        (NS, "Scenario"), "realized_scenarios"
    )
    constraint_durations = m.Containment["ConstraintDuration"](
        "ownedConstraintDurations", (NS, "ConstraintDuration")
    )

    @property
    def related_functions(self) -> m.ElementList[fa.AbstractFunction]:
        return self.fragments.map("function")


class InteractionFragment(capellacore.NamedElement, abstract=True):
    covered_instance_roles = m.Association["InstanceRole"](
        (NS, "InstanceRole"), "coveredInstanceRoles", legacy_by_type=True
    )

    if not t.TYPE_CHECKING:
        covered = m.DeprecatedAccessor("covered_instance_roles")


class AbstractEnd(InteractionFragment, abstract=True):
    event = m.Single["Event"](m.Association((NS, "Event"), "event"))


class MessageEnd(AbstractEnd):
    pass


class TimeLapse(capellacore.NamedElement, abstract=True):
    start = m.Single["InteractionFragment"](
        m.Association((NS, "InteractionFragment"), "start")
    )
    finish = m.Single["InteractionFragment"](
        m.Association((NS, "InteractionFragment"), "finish")
    )


class Execution(TimeLapse):
    pass


class ExecutionEnd(AbstractEnd):
    pass


class Event(capellacore.NamedElement, behavior.AbstractEvent, abstract=True):
    pass


class CreationEvent(Event):
    pass


class DestructionEvent(Event):
    pass


class ExecutionEvent(Event):
    pass


class InstanceRole(capellacore.NamedElement):
    represented_instance = m.Single["information.AbstractInstance"](
        m.Association(
            (ns.INFORMATION, "AbstractInstance"), "representedInstance"
        )
    )

    if not t.TYPE_CHECKING:
        instance = m.DeprecatedAccessor("represented_instance")


class _EventOperation(Event):
    # TODO delete this class when removing deprecated features
    operation = m.Association["information.AbstractEventOperation"](
        (ns.INFORMATION, "AbstractEventOperation"), "operation"
    )


class EventReceiptOperation(_EventOperation):
    operation = m.Association["information.AbstractEventOperation"](
        (ns.INFORMATION, "AbstractEventOperation"), "operation"
    )


class EventSentOperation(_EventOperation):
    operation = m.Association["information.AbstractEventOperation"](
        (ns.INFORMATION, "AbstractEventOperation"), "operation"
    )


class MergeLink(capellacore.Trace):
    pass


class RefinementLink(capellacore.Trace):
    pass


class AbstractCapabilityRealization(capellacore.Allocation):
    pass


class AbstractCapability(
    capellacore.Structure,
    capellacore.InvolverElement,
    fa.AbstractFunctionalChainContainer,
    abstract=True,
):
    precondition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "preCondition")
    )
    postcondition = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "postCondition")
    )
    scenarios = m.Containment["Scenario"]("ownedScenarios", (NS, "Scenario"))
    extends = m.Containment["AbstractCapabilityExtend"](
        "extends", (NS, "AbstractCapabilityExtend")
    )
    extended_by = m.Backref["AbstractCapabilityExtend"](
        (NS, "AbstractCapabilityExtend"), "target"
    )
    extension_points = m.Containment["AbstractCapabilityExtensionPoint"](
        "abstractCapabilityExtensionPoints",
        (NS, "AbstractCapabilityExtensionPoint"),
    )
    generalizations = m.Containment["AbstractCapabilityGeneralization"](
        "superGeneralizations", (NS, "AbstractCapabilityGeneralization")
    )
    generalized_by = m.Backref["AbstractCapability"](
        (NS, "AbstractCapability"), "target"
    )
    includes = m.Containment["AbstractCapabilityInclude"](
        "includes", (NS, "AbstractCapabilityInclude")
    )
    included_by = m.Backref["AbstractCapabilityInclude"](
        (NS, "AbstractCapabilityInclude"), "included"
    )
    chain_involvements = m.Containment[
        "FunctionalChainAbstractCapabilityInvolvement"
    ](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        (NS, "FunctionalChainAbstractCapabilityInvolvement"),
    )
    involved_chains = m.Allocation["fa.FunctionalChain"](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        (NS, "FunctionalChainAbstractCapabilityInvolvement"),
        (ns.FA, "AbstractFunction"),
        attr="involved",
    )
    function_involvements = m.Containment[
        "AbstractFunctionAbstractCapabilityInvolvement"
    ](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        (NS, "AbstractFunctionAbstractCapabilityInvolvement"),
    )
    involved_functions = m.Allocation["fa.AbstractFunction"](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        (NS, "AbstractFunctionAbstractCapabilityInvolvement"),
        (ns.FA, "AbstractFunction"),
        attr="involved",
    )
    available_in_states = m.Association["capellacommon.State"](
        (ns.CAPELLACOMMON, "State"), "availableInStates"
    )
    capability_realizations = m.Containment["AbstractCapabilityRealization"](
        "ownedAbstractCapabilityRealizations",
        (NS, "AbstractCapabilityRealization"),
    )
    realized_capabilities = m.Allocation["AbstractCapability"](
        "ownedAbstractCapabilityRealizations",
        (NS, "AbstractCapabilityRealization"),
        (NS, "AbstractCapability"),
        attr="targetElement",
    )
    realizing_capabilities = m.Backref["AbstractCapability"](
        (NS, "AbstractCapability"), "realized_capabilities"
    )

    if not t.TYPE_CHECKING:
        generalizes = m.DeprecatedAccessor("generalizations")
        states = m.DeprecatedAccessor("available_in_states")


class AbstractCapabilityExtend(capellacore.Relationship):
    _xmltag = "extends"

    source = m.Alias["m.ModelElement"]("parent")
    extended = m.Single["AbstractCapability"](
        m.Association((NS, "AbstractCapability"), "extended")
    )
    target = m.Alias["AbstractCapability"]("extended")
    extension_location = m.Single["AbstractCapabilityExtensionPoint"](
        m.Association(
            (NS, "AbstractCapabilityExtensionPoint"), "extensionLocation"
        )
    )


class AbstractCapabilityExtensionPoint(capellacore.NamedRelationship):
    extend_links = m.Association["AbstractCapabilityExtend"](
        (NS, "AbstractCapabilityExtend"), "extendLinks"
    )


class AbstractCapabilityGeneralization(capellacore.Relationship):
    _xmltag = "superGeneralizations"

    source = m.Alias["m.ModelElement"]("parent")
    super = m.Single["AbstractCapability"](
        m.Association((NS, "AbstractCapability"), "super")
    )
    target = m.Alias["AbstractCapability"]("super")


class AbstractCapabilityInclude(capellacore.Relationship):
    _xmltag = "includes"

    source = m.Alias["m.ModelElement"]("parent")
    included = m.Single["AbstractCapability"](
        m.Association((NS, "AbstractCapability"), "included")
    )
    target = m.Alias["AbstractCapability"]("included")


class InteractionState(InteractionFragment):
    state = m.Single["capellacommon.AbstractState"](
        m.Association(
            (ns.CAPELLACOMMON, "AbstractState"), "relatedAbstractState"
        )
    )
    function = m.Single["fa.AbstractFunction"](
        m.Association((ns.FA, "AbstractFunction"), "relatedAbstractFunction")
    )


class AbstractFragment(TimeLapse, abstract=True):
    gates = m.Containment["Gate"]("ownedGates", (NS, "Gate"))


class InteractionUse(AbstractFragment):
    referenced_scenario = m.Single["Scenario"](
        m.Association((NS, "Scenario"), "referencedScenario")
    )


class CombinedFragment(AbstractFragment):
    operator = m.EnumPOD("operator", InteractionOperatorKind)
    operands = m.Association["InteractionOperand"](
        (NS, "InteractionOperand"), "referencedOperands"
    )


class Gate(MessageEnd):
    pass


class InteractionOperand(InteractionFragment):
    fragments = m.Association["InteractionFragment"](
        (NS, "InteractionFragment"), "referencedInteractionFragments"
    )
    guard = m.Single["capellacore.Constraint"](
        m.Association((ns.CAPELLACORE, "Constraint"), "guard")
    )


class FragmentEnd(InteractionFragment):
    pass


class FunctionalChainAbstractCapabilityInvolvement(capellacore.Involvement):
    pass


class AbstractFunctionAbstractCapabilityInvolvement(capellacore.Involvement):
    pass


class ScenarioRealization(capellacore.Allocation):
    pass


class StateFragment(TimeLapse):
    state = m.Single["capellacommon.AbstractState"](
        m.Association(
            (ns.CAPELLACOMMON, "AbstractState"), "relatedAbstractState"
        )
    )
    function = m.Single["fa.AbstractFunction"](
        m.Association((ns.FA, "AbstractFunction"), "relatedAbstractFunction")
    )


class ArmTimerEvent(Event):
    pass


class CancelTimerEvent(Event):
    pass


class ConstraintDuration(capellacore.NamedElement):
    duration = m.StringPOD("duration")
    start = m.Single["InteractionFragment"](
        m.Association((NS, "InteractionFragment"), "start")
    )
    finish = m.Single["InteractionFragment"](
        m.Association((NS, "InteractionFragment"), "finish")
    )


class SequenceMessageValuation(capellacore.CapellaElement):
    element = m.Single["information.ExchangeItemElement"](
        m.Association(
            (ns.INFORMATION, "ExchangeItemElement"), "exchangeItemElement"
        )
    )
    value = m.Single["modellingcore.ValueSpecification"](
        m.Association((ns.MODELLINGCORE, "ValueSpecification"), "value")
    )


if not t.TYPE_CHECKING:

    def __getattr__(name):
        if name == "Exchange":
            warnings.warn(
                "The Exchange class has been removed, use capellacore.Relationship instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return capellacore.Relationship

        if name == "AbstractInvolvement":
            warnings.warn(
                "interaction.AbstractInvolvement has been moved to capellacore.Involvement",
                DeprecationWarning,
                stacklevel=2,
            )
            return capellacore.Involvement

        if name == "EventOperation":
            warnings.warn(
                (
                    "EventOperation has been removed,"
                    " use EventSentOperation/EventReceiptOperation directly"
                    " or another common super class instead"
                ),
                DeprecationWarning,
                stacklevel=2,
            )
            return _EventOperation

        raise AttributeError(name)


from . import capellacommon, information, modellingcore  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import pathlib

import pytest
from click import testing as clitest

import capellambse.metamodel as mm
import capellambse.model as m
from capellambse._scripts import validate as cli_main
from capellambse.extensions import validation

TEST_UUID = "da12377b-fb70-4441-8faa-3a5c153c5de2"
TEST_RULE_ID = "Rule-001"
TEST_RULE_PARAMS = {
    "name": "Always True",
    "rationale": "True",
    "action": "Nothing",
}


@pytest.fixture
def fake_registry(monkeypatch):
    registry = validation.Rules()
    monkeypatch.setattr(validation._validate, "_VALIDATION_RULES", registry)
    return registry


def test_decorated_rules_are_added_to_global_registry(fake_registry):
    @validation.rule(
        TEST_RULE_ID,
        validation.Category.REQUIRED,
        types=[mm.la.LogicalFunction],
        **TEST_RULE_PARAMS,
    )
    def testrule(_):
        return True

    assert list(fake_registry.items()) == [(testrule.id, testrule)]


def test_rules_can_be_filtered_by_object_type(fake_registry):
    @validation.rule(
        TEST_RULE_ID,
        validation.Category.REQUIRED,
        types=[mm.la.LogicalComponent, mm.la.LogicalFunction],
        **TEST_RULE_PARAMS,
    )
    def testrule(_):
        return True

    assert list(fake_registry.items()) == [(testrule.id, testrule)]
    assert fake_registry.by_class(mm.la.LogicalComponent) == [testrule]
    assert fake_registry.by_class(mm.la.LogicalFunction) == [testrule]
    assert fake_registry.by_class(mm.sa.SystemFunction) == []


def test_model_gives_access_to_the_full_set_of_rules(
    model: m.MelodyModel,
) -> None:
    assert len(model.validation.rules) > 0


def test_model_object_gives_access_to_rules_that_apply_to_it(
    model: m.MelodyModel,
):
    obj = model.by_uuid(TEST_UUID)
    assert isinstance(obj.validation, validation.ElementValidation)

    rules = obj.validation.rules

    assert len(rules) > 0
    assert all(rule.applies_to(obj) for rule in rules)


@pytest.mark.usefixtures("fake_registry")
def test_validation_results_filtering(model: m.MelodyModel) -> None:
    @validation.rule(
        "TEST-LC",
        validation.Category.REQUIRED,
        types=[mm.la.LogicalComponent],
        **TEST_RULE_PARAMS,
    )
    def rule_lc(_: m.ModelElement) -> bool:
        return True

    @validation.rule(
        "TEST-LF",
        validation.Category.RECOMMENDED,
        types=[mm.la.LogicalFunction],
        **TEST_RULE_PARAMS,
    )
    def rule_lf(_: m.ModelElement) -> bool:
        return True

    assert model.search(mm.la.LogicalComponent), "Empty test model?"
    assert model.search(mm.la.LogicalFunction), "Empty test model?"

    results = model.validation.validate()
    assert results

    required = results.by_category("REQUIRED")
    assert len(required) > 0
    assert all(i.category.name == "REQUIRED" for i in required.iter_rules())

    component = results.by_class("LogicalComponent")
    assert len(component) > 0
    assert sum(1 for _ in component.iter_objects()) == len(
        model.search("LogicalComponent")
    )
    assert all(
        type(i).__name__ == "LogicalComponent"
        for i in component.iter_objects()
    )

    function = results.by_class("LogicalFunction")
    assert len(function) > 0
    assert (
        sum(1 for _ in function.iter_objects())
        # -1 because the root function is not validated
        == len(model.search("LogicalFunction")) - 1
    )
    assert all(
        type(i).__name__ == "LogicalFunction" for i in function.iter_objects()
    )


def test_MelodyModel_validation(model: m.MelodyModel):
    assert isinstance(model.validation, validation.ModelValidation)
    assert model.validation.rules

    results = model.validate()

    assert results


def test_ModelObject_validation(model: m.MelodyModel):
    obj = model.by_uuid(TEST_UUID)
    assert isinstance(obj.validation, validation.ElementValidation)
    assert obj.validation.rules

    results = obj.validate()

    assert results


def test_cli_creates_a_validation_report(tmp_path: pathlib.Path):
    output_file = tmp_path / "report.html"
    runner = clitest.CliRunner()

    result = runner.invoke(
        cli_main.main,
        ["-mtest-7.0", f"-o{output_file}"],
    )

    assert result.exit_code == 0, f"CLI returned code {result.exit_code}"
    assert output_file.exists()
    assert output_file.read_bytes() != b""



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum
import typing as t

import capellambse.model as m

from .. import capellacore, modellingcore
from .. import namespaces as ns
from . import datavalue

NS = ns.INFORMATION_DATATYPE
NS_DV = ns.INFORMATION_DATAVALUE


@m.stringy_enum
@enum.unique
class NumericTypeKind(enum.Enum):
    """The kind of this numeric data type."""

    INTEGER = "INTEGER"
    FLOAT = "FLOAT"


class DataType(
    capellacore.GeneralizableElement,
    datavalue.DataValueContainer,
    modellingcore.FinalizableElement,
    abstract=True,
):
    _xmltag = "ownedDataTypes"

    is_discrete = m.BoolPOD("discrete")
    """Whether or not this data type characterizes a discrete value."""
    is_min_inclusive = m.BoolPOD("minInclusive")
    is_max_inclusive = m.BoolPOD("maxInclusive")
    pattern = m.StringPOD("pattern")
    """Textual specification of a constraint associated to this data type."""
    visibility = m.EnumPOD("visibility", capellacore.VisibilityKind)
    information_realizations = m.Containment["InformationRealization"](
        "ownedInformationRealizations",
        (ns.INFORMATION, "InformationRealization"),
    )

    if not t.TYPE_CHECKING:
        min_inclusive = m.DeprecatedAccessor("is_min_inclusive")
        max_inclusive = m.DeprecatedAccessor("is_max_inclusive")


class BooleanType(DataType):
    literals = m.Containment["datavalue.LiteralBooleanValue"](
        "ownedLiterals", (NS_DV, "LiteralBooleanValue"), fixed_length=2
    )
    default_value = m.Single["datavalue.AbstractBooleanValue"](
        m.Containment("ownedDefaultValue", (NS_DV, "AbstractBooleanValue"))
    )
    if not t.TYPE_CHECKING:
        default = m.DeprecatedAccessor("default_value")


class Enumeration(DataType):
    owned_literals = m.Containment["datavalue.EnumerationLiteral"](
        "ownedLiterals", (NS_DV, "EnumerationLiteral")
    )
    default_value = m.Single["datavalue.AbstractEnumerationValue"](
        m.Containment("ownedDefaultValue", (NS_DV, "AbstractEnumerationValue"))
    )
    null_value = m.Single["datavalue.AbstractEnumerationValue"](
        m.Containment("ownedNullValue", (NS_DV, "AbstractEnumerationValue"))
    )
    min_value = m.Single["datavalue.AbstractEnumerationValue"](
        m.Containment("ownedMinValue", (NS_DV, "AbstractEnumerationValue"))
    )
    max_value = m.Single["datavalue.AbstractEnumerationValue"](
        m.Containment("ownedMaxValue", (NS_DV, "AbstractEnumerationValue"))
    )
    domain_type = m.Single["DataType"](
        m.Association((NS, "DataType"), "domainType")
    )

    @property
    def literals(self) -> m.ElementList[datavalue.EnumerationLiteral]:
        """Return all owned and inherited literals."""
        return (
            self.owned_literals + self.super.literals
            if isinstance(self.super, Enumeration)
            else self.owned_literals
        )


class StringType(DataType):
    default_value = m.Single["datavalue.AbstractStringValue"](
        m.Containment("ownedDefaultValue", (NS_DV, "AbstractStringValue"))
    )
    null_value = m.Single["datavalue.AbstractStringValue"](
        m.Containment("ownedNullValue", (NS_DV, "AbstractStringValue"))
    )
    min_length = m.Single["datavalue.NumericValue"](
        m.Containment("ownedMinLength", (NS_DV, "NumericValue"))
    )
    max_length = m.Single["datavalue.NumericValue"](
        m.Containment("ownedMaxLength", (NS_DV, "NumericValue"))
    )


class NumericType(DataType):
    kind = m.EnumPOD("kind", NumericTypeKind)
    default_value = m.Single["datavalue.NumericValue"](
        m.Containment("ownedDefaultValue", (NS_DV, "NumericValue"))
    )
    null_value = m.Single["datavalue.NumericValue"](
        m.Containment("ownedNullValue", (NS_DV, "NumericValue"))
    )
    min_value = m.Single["datavalue.NumericValue"](
        m.Containment("ownedMinValue", (NS_DV, "NumericValue"))
    )
    max_value = m.Single["datavalue.NumericValue"](
        m.Containment("ownedMaxValue", (NS_DV, "NumericValue"))
    )


class PhysicalQuantity(NumericType):
    unit = m.Single["Unit"](m.Association((NS, "Unit"), "unit"))


from . import InformationRealization, Unit  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

__all__ = [
    "CapellaIncomingRelation",
    "CapellaModule",
    "CapellaOutgoingRelation",
    "CapellaRelation",
    "CapellaTypesFolder",
    "ElementRelationAccessor",
    "RelationsList",
    "RequirementsRelationAccessor",
]

import collections.abc as cabc
import contextlib
import os
import sys
import typing as t
import warnings

from lxml import etree

import capellambse
import capellambse.model as m
from capellambse.metamodel import namespaces as ns

from . import exporter, requirements

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

if t.TYPE_CHECKING:
    from capellambse.metamodel import capellacore  # noqa: F401

NS = m.Namespace(
    "http://www.polarsys.org/capella/requirements",
    "CapellaRequirements",
    "org.polarsys.capella.vp.requirements",
)


class CapellaTypesFolder(requirements.TypesFolder):
    _xmltag = "ownedExtensions"

    @property
    @deprecated(
        (
            "CapellaTypesFolder.module_types is deprecated,"
            " use 'types.by_class(ModuleType)' instead"
        ),
        category=FutureWarning,
    )
    def module_types(self) -> m.ElementList[requirements.ModuleType]:
        return self.types.by_class(requirements.ModuleType)

    @property
    @deprecated(
        (
            "CapellaTypesFolder.relation_types is deprecated,"
            " use 'types.by_class(ModuleType)' instead"
        ),
        category=FutureWarning,
    )
    def relation_types(self) -> m.ElementList[requirements.RelationType]:
        return self.types.by_class(requirements.RelationType)

    @property
    @deprecated(
        (
            "CapellaTypesFolder.requirement_types is deprecated,"
            " use 'types.by_class(ModuleType)' instead"
        ),
        category=FutureWarning,
    )
    def requirement_types(self) -> m.ElementList[requirements.RequirementType]:
        return self.types.by_class(requirements.RequirementType)

    if not t.TYPE_CHECKING:
        data_type_definitions = m.DeprecatedAccessor("definition_types")


class CapellaModule(requirements.Module):
    _xmltag = "ownedExtensions"

    requirement_types_folders = m.Filter["CapellaTypesFolder"](
        "extensions", (NS, "CapellaTypesFolder")
    )

    @property
    @deprecated(
        (
            "CapellaModule.folders is deprecated,"
            " use 'requirements.by_class(Folder)' instead"
        ),
        category=FutureWarning,
    )
    def folders(self) -> m.ElementList[requirements.Folder]:
        return self.requirements.by_class(requirements.Folder)

    def to_reqif(
        self,
        to: str | os.PathLike | t.IO[bytes],
        *,
        metadata: cabc.Mapping[str, t.Any] | None = None,
        pretty: bool = False,
        compress: bool | None = None,
    ) -> None:
        """Export this module as ReqIF XML.

        You can override some auto-generated metadata placed in the header
        section by passing a dictionary as ``metadata``. The following keys
        are supported. Unsupported keys are silently ignored.

        -   ``creation_time``: A ``datetime.datetime`` object that specifies
            this document's creation time. Default to the current time.
        -   ``comment``: Override the ReqIF file's comment. Defaults to a
            text derived from the model and module names.
        -   ``title``: Specify the document title. Defaults to the module's
            ``long_name``.

        Parameters
        ----------
        to
            Where to export to. Can be the name of a file, or a file-like
            object opened in binary mode.
        metadata
            A dictionary with additional metadata (see above).
        pretty
            Format the XML human-readable.
        compress
            Write compressed data (``*.reqifz``). Defaults to ``True``
            if ``target`` is a string or path-like and its name ends in
            ``.reqifz``, otherwise defaults to ``False``.
        """
        exporter.export_module(
            self, to, metadata=metadata, pretty=pretty, compress=compress
        )


class CapellaRelation(requirements.AbstractRelation, abstract=True):
    pass


class CapellaIncomingRelation(CapellaRelation):
    _xmltag = "ownedRelations"

    def __init__(
        self,
        model: m.MelodyModel,
        parent: etree._Element,
        xmltag: str | None = None,
        /,
        **kw: t.Any,
    ) -> None:
        kw.setdefault("source", m.wrap_xml(model, parent))
        super().__init__(model, parent, xmltag, **kw)

    source = m.Single["requirements.Requirement"](
        m.Association((requirements.NS, "Requirement"), "source")
    )
    target = m.Single["capellacore.CapellaElement"](
        m.Association((ns.CAPELLACORE, "CapellaElement"), "target")
    )


class CapellaOutgoingRelation(CapellaRelation):
    _xmltag = "ownedExtensions"

    def __init__(
        self,
        model: m.MelodyModel,
        parent: etree._Element,
        xmltag: str | None = None,
        /,
        **kw: t.Any,
    ) -> None:
        kw.setdefault("target", m.wrap_xml(model, parent))
        super().__init__(model, parent, xmltag, **kw)

    # NOTE: source/target are swapped intentionally here,
    # so that 'some_relation.source' is always a Requirement
    target = m.Single["capellacore.CapellaElement"](
        m.Association((ns.CAPELLACORE, "CapellaElement"), "source")
    )
    source = m.Single["requirements.Requirement"](
        m.Association((requirements.NS, "Requirement"), "target")
    )


class RequirementsRelationAccessor(
    m.WritableAccessor[requirements.AbstractRelation]
):
    """Searches for requirement relations in the architecture layer."""

    __slots__ = ("aslist",)

    def __init__(self, *args, **kw) -> None:
        super().__init__(
            *args, **kw, aslist=m.ElementList, single_attr="long_name"
        )

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        return self._make_list(obj, self._find_relations(obj))

    def __set__(self, obj, value: t.Any) -> None:
        if not isinstance(value, cabc.Iterable):
            value = (value,)

        if CapellaOutgoingRelation in [type(i) for i in value]:
            raise NotImplementedError("Cannot insert CapellaOutgoingRelations")

        for i in self._find_relations(obj):
            ip = i.getparent()
            assert ip is not None
            ip.remove(i)
        obj._element.extend(value)

    def __delete__(self, obj) -> None:
        assert self.aslist is not None

        for i in self._find_relations(obj):
            parent = i.getparent()
            assert parent is not None
            parent.remove(i)

    def _find_relations(self, obj) -> list[etree._Element]:
        rels = obj._model.search(
            CapellaIncomingRelation,
            requirements.InternalRelation,
            CapellaOutgoingRelation,
        )
        return [i._element for i in rels if obj in (i.source, i.target)]

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model, elements, m.ModelElement, parent=parent_obj
        )

    def create(
        self,
        elmlist: m.ElementListCouplingMixin,
        typehint: str | None = None,
        /,
        **kw: t.Any,
    ) -> requirements.InternalRelation | CapellaIncomingRelation:
        del typehint

        warnings.warn(
            (
                "Mutating Requirement.relations directly is deprecated"
                " and will soon no longer affect the model,"
                " mutate the respective Containment feature instead"
                " (e.g. Requirement.owned_relations)"
            ),
            category=FutureWarning,
            stacklevel=2,
        )

        if "target" not in kw:
            raise m.InvalidModificationError(
                "No `target` for new requirement relation"
            )
        cls = self._find_relation_type(kw["target"])
        parent = elmlist._parent._element
        with elmlist._model._loader.new_uuid(parent) as uuid:
            return cls(
                elmlist._model,
                parent,
                **kw,
                source=elmlist._parent,
                uuid=uuid,
            )

    def delete(self, elmlist, obj) -> None:
        warnings.warn(
            (
                "Mutating Requirement.relations directly is deprecated"
                " and will soon no longer affect the model,"
                " mutate the respective Containment feature instead"
                " (e.g. Requirement.owned_relations)"
            ),
            category=FutureWarning,
            stacklevel=2,
        )

        assert self.aslist is not None
        relations = self._find_relations(elmlist._parent)
        for relation in relations:
            if relation == obj._element:
                obj.parent._model._loader.idcache_remove(relation)
                obj.parent._element.remove(relation)
                break
        else:
            raise ValueError("Cannot delete: Target object not in this list")

    def insert(
        self,
        elmlist: m.ElementListCouplingMixin,
        index: int,
        value: m.ModelObject | m.NewObject,
    ) -> None:
        warnings.warn(
            (
                "Mutating Requirement.relations directly is deprecated"
                " and will soon no longer affect the model,"
                " mutate the respective Containment feature instead"
                " (e.g. Requirement.owned_relations)"
            ),
            category=FutureWarning,
            stacklevel=2,
        )

        if isinstance(value, m.NewObject):
            raise NotImplementedError("Cannot insert new objects yet")

        if isinstance(value, CapellaOutgoingRelation):
            if not value.target:
                raise RuntimeError(
                    "Cannot insert outgoing relation without target:"
                    f" {value._short_repr_()}"
                )
            parent = value.target._element
        else:
            assert isinstance(
                value, CapellaIncomingRelation | requirements.InternalRelation
            )
            assert elmlist._parent == value.source
            parent = elmlist._parent._element

        parent.insert(index, value._element)
        elmlist._model._loader.idcache_index(value._element)

    @contextlib.contextmanager
    def purge_references(
        self, obj: m.ModelObject, target: m.ModelObject
    ) -> cabc.Generator[None, None, None]:
        """Do nothing.

        This is a no-op, as this accessor provides a virtual relation.

        The relation objects it handles are cleaned up by removing the
        source or target attribute.
        """
        del obj, target
        yield

    def _find_relation_type(
        self, target: m.ModelElement
    ) -> type[requirements.InternalRelation | CapellaIncomingRelation]:
        if isinstance(target, requirements.Requirement):
            return requirements.InternalRelation
        if isinstance(target, requirements.ReqIFElement):
            raise m.InvalidModificationError(
                "Cannot create relations to targets of type"
                f" {type(target).__name__}"
            )
        return CapellaIncomingRelation


class RelationsList(m.ElementList[requirements.AbstractRelation]):
    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[t.Any] | None = None,
        *,
        source: m.ModelObject,
    ) -> None:
        del elemclass
        super().__init__(model, elements)
        self._source = source

    @t.overload
    def __getitem__(self, idx: int) -> requirements.AbstractRelation: ...
    @t.overload
    def __getitem__(self, idx: slice) -> RelationsList: ...
    @t.overload
    def __getitem__(self, idx: str) -> t.Any: ...
    def __getitem__(self, idx: int | slice | str) -> t.Any:
        rel = super().__getitem__(idx)
        if isinstance(rel, m.ElementList):
            return rel

        assert isinstance(rel, requirements.AbstractRelation)
        assert self._source in (rel.source, rel.target)
        if self._source == rel.source:
            return rel.target
        return rel.source

    def by_relation_type(self, reltype: str) -> RelationsList:
        warnings.warn(
            (
                "Requirement.related.by_relation_type is deprecated,"
                "use 'Requirement.relations.by_type(...)'"
                " or 'ModelElement.requirements_relations.by_type(...)' instead"
            ),
            category=FutureWarning,
            stacklevel=2,
        )

        matches = []
        for elm in self._elements:
            rel_elm = m.wrap_xml(self._model, elm)
            assert isinstance(rel_elm, requirements.AbstractRelation)
            if rel_elm.type is not None and rel_elm.type.long_name == reltype:
                matches.append(elm)
        return self._newlist(matches)

    def by_relation_class(
        self, class_: t.Literal["incoming", "outgoing", "internal"]
    ) -> RelationsList:
        warnings.warn(
            (
                "Requirement.related.by_relation_class is deprecated,"
                "use 'Requirement.relations.by_class(...)' instead"
            ),
            category=FutureWarning,
            stacklevel=2,
        )

        relation_types = {
            "incoming": CapellaIncomingRelation,
            "outgoing": CapellaOutgoingRelation,
            "internal": requirements.InternalRelation,
        }
        matches: list[etree._Element] = []
        for elm in self._elements:
            rel_elm = m.wrap_xml(self._model, elm)
            if isinstance(rel_elm, relation_types[class_]):
                matches.append(rel_elm._element)
        return self._newlist(matches)

    def _newlist(self, elements: list[etree._Element]) -> RelationsList:
        listtype = self._newlist_type()
        assert issubclass(listtype, RelationsList)
        return listtype(self._model, elements, source=self._source)


class ElementRelationAccessor(m.WritableAccessor[m.ModelElement]):
    """Provides access to RequirementsRelations of a ModelElement."""

    __slots__ = ("aslist",)

    def __init__(self) -> None:
        super().__init__(aslist=RelationsList, single_attr="long_name")

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        relations: list[etree._Element] = []
        for relation in obj._model.search(
            CapellaIncomingRelation,
            requirements.InternalRelation,
            CapellaOutgoingRelation,
        ):
            if None in (relation.source, relation.target):
                continue
            if obj in (relation.source, relation.target):
                relations.append(relation._element)
        return self._make_list(obj, relations)

    @contextlib.contextmanager
    def purge_references(
        self, obj: m.ModelObject, target: m.ModelObject
    ) -> cabc.Generator[None, None, None]:
        """Do nothing.

        This is a no-op, as this accessor provides a virtual relation.

        The relation objects it handles are cleaned up by removing the
        source or target attribute.
        """
        del obj, target
        yield

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model,
            elements,
            None,
            parent=parent_obj,
            source=parent_obj,
        )


requirements.Requirement.relations = RequirementsRelationAccessor()  # type: ignore[deprecated]
requirements.Requirement.related = ElementRelationAccessor()  # type: ignore[deprecated]



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Main entry point for various CLI scripts around capellambse."""

import contextlib
import importlib
import importlib.resources as imr

import click

from . import _scripts


class LazyGroup(click.Group):
    def list_commands(self, ctx):
        cmds: list[str] = []
        for i in imr.files(_scripts).iterdir():
            if i.name.endswith(".py") and not i.name.startswith("_"):
                cmds.append(i.name.removesuffix(".py").replace("_", "-"))
        cmds.sort()
        return super().list_commands(ctx) + cmds

    def get_command(self, ctx, name):
        with contextlib.suppress(ImportError):
            modname = name.replace("-", "_")
            cmd = importlib.import_module(
                f"{_scripts.__name__}.{modname}"
            ).main
            assert isinstance(cmd, click.Command)
            cmd.name = name
            return cmd

        return super().get_command(ctx, name)


@click.group(cls=LazyGroup, no_args_is_help=True)
def main():
    pass


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

import capellambse

# -- Project information -----------------------------------------------------

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
with open("../../pyproject.toml", "rb") as f:
    _metadata = tomllib.load(f)["project"]

project = "py-capellambse"
pypi = "capellambse"
author = _metadata["authors"][0]["name"]
copyright = f"{author} and the {_metadata['name']} contributors"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_click",
    "sphinx_autodoc_typehints",
]

nitpicky = True
nitpick_ignore = [
    # in inventory as "py:function"
    ("py:class", "operator.attrgetter"),
    # exposed as `json.JSONEncoder`
    ("py:class", "json.encoder.JSONEncoder"),
    # TODO figure out why sphinx doesn't document these
    ("py:exc", "capellambse.UnsupportedPluginError"),
    ("py:exc", "capellambse.UnsupportedPluginVersionError"),
    ("any", "capellambse.aird.GLOBAL_FILTERS"),
    # Private type hinting helpers
    ("py:class", "_MapFunction"),
    ("py:class", "_NotSpecifiedType"),
    ("py:class", "capellambse.model._descriptors._Specification"),
    # Deprecated ABC
    ("py:class", "capellambse.metamodel.fa._AbstractExchange"),
    ("py:class", "capellambse.metamodel.interaction._CapabilityRelation"),
    ("py:class", "capellambse.metamodel.interaction._EventOperation"),
]
nitpick_ignore_regex = [
    ("py:.*", r"^(?:awesomeversion|yaml)\..*"),
    ("py:.*", r"^(?:.*\.)?_[A-Z]$"),  # Single-letter TypeVars (e.g. _T)
    ("py:.*", r"^(?:.*\.)?(?:_UnspecifiedType|_NOT_SPECIFIED)$"),
    # Super/subclass and "see also" references sometimes break
    ("py:(meth|obj)", r"(?:.*\.)?write_transaction"),
    # Sometimes autodoc_typehints doesn't properly resolve aliases
    ("py:class", r"(?:cabc|etree|m|t|_obj)\..*"),
]


# Add any paths that contain templates here, relative to this directory.
templates_path: list[str] = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: list[str] = []


# -- General information about the project -----------------------------------

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The full version, including alpha/beta/rc tags.
version = capellambse.__version__
rst_epilog = f"""
.. |Project| replace:: {project}
.. |Version| replace:: {version}
"""


# -- Options for auto-doc ----------------------------------------------------
autoclass_content = "class"
autodoc_class_signature = "separated"
autodoc_typehints = "description"

autodoc_type_aliases = {
    "capellambse.model.common.accessors._NewObject": "capellambse.model.new_object",
}


# -- Options for napoleon ----------------------------------------------------
napoleon_custom_sections = ["Transaction arguments", "Well-known arguments"]
napoleon_google_docstring = False
napoleon_include_init_with_doc = True


# -- Options for Intersphinx output ------------------------------------------
intersphinx_mapping = {
    "click": ("https://click.palletsprojects.com/en/latest", None),
    "lxml": ("https://lxml.de/apidoc/", None),
    "markupsafe": ("https://markupsafe.palletsprojects.com/en/2.1.x", None),
    "python": ("https://docs.python.org/3", None),
    "svgwrite": ("https://svgwrite.readthedocs.io/en/latest", None),
    "requests": ("https://requests.readthedocs.io/en/latest", None),
    "PIL": ("https://pillow.readthedocs.io/en/stable", None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation
# for a list of builtin themes.

html_theme = "furo"
html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/dbinfrago/py-capellambse",
            "html": '<img src="data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHN0cm9rZT0iY3VycmVudENvbG9yIiBmaWxsPSJjdXJyZW50Q29sb3IiIHN0cm9rZS13aWR0aD0iMCIgdmlld0JveD0iMCAwIDE2IDE2Ij48cGF0aCBmaWxsLXJ1bGU9ImV2ZW5vZGQiIGQ9Ik04IDBDMy41OCAwIDAgMy41OCAwIDhjMCAzLjU0IDIuMjkgNi41MyA1LjQ3IDcuNTkuNC4wNy41NS0uMTcuNTUtLjM4IDAtLjE5LS4wMS0uODItLjAxLTEuNDktMi4wMS4zNy0yLjUzLS40OS0yLjY5LS45NC0uMDktLjIzLS40OC0uOTQtLjgyLTEuMTMtLjI4LS4xNS0uNjgtLjUyLS4wMS0uNTMuNjMtLjAxIDEuMDguNTggMS4yMy44Mi43MiAxLjIxIDEuODcuODcgMi4zMy42Ni4wNy0uNTIuMjgtLjg3LjUxLTEuMDctMS43OC0uMi0zLjY0LS44OS0zLjY0LTMuOTUgMC0uODcuMzEtMS41OS44Mi0yLjE1LS4wOC0uMi0uMzYtMS4wMi4wOC0yLjEyIDAgMCAuNjctLjIxIDIuMi44Mi42NC0uMTggMS4zMi0uMjcgMi0uMjcuNjggMCAxLjM2LjA5IDIgLjI3IDEuNTMtMS4wNCAyLjItLjgyIDIuMi0uODIuNDQgMS4xLjE2IDEuOTIuMDggMi4xMi41MS41Ni44MiAxLjI3LjgyIDIuMTUgMCAzLjA3LTEuODcgMy43NS0zLjY1IDMuOTUuMjkuMjUuNTQuNzMuNTQgMS40OCAwIDEuMDctLjAxIDEuOTMtLjAxIDIuMiAwIC4yMS4xNS40Ni41NS4zOEE4LjAxMyA4LjAxMyAwIDAgMCAxNiA4YzAtNC40Mi0zLjU4LTgtOC04eiIvPjwvc3ZnPg=="/>',
            "class": "",
        },
    ],
}
html_short_title = "py-capellambse"
html_show_sourcelink = False

# -- Extra options for Furo theme --------------------------------------------

pygments_style = "tango"
pygments_dark_style = "monokai"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Skip __new__ methods ----------------------------------------------------
# This skips all __new__ methods. They only appear for NamedTuple
# classes, and they don't show any useful information that's not
# documented on the members already anyways.
def skip_dunder_new(app, what, name, obj, skip, options) -> bool:
    del app, obj, options
    return skip or (what == "class" and name == "__new__")


def setup(app):
    app.connect("autodoc-skip-member", skip_dunder_new)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import collections.abc as cabc
import dataclasses
import errno
import hashlib
import logging
import os
import pathlib
import re
import shlex
import shutil
import subprocess
import tempfile
import textwrap
import typing as t
import urllib.parse
import weakref

import capellambse.helpers

from . import abc

LOGGER = logging.getLogger(__name__)
CACHEBASE = pathlib.Path(capellambse.dirs.user_cache_dir, "models")
WTBASE = pathlib.Path(capellambse.dirs.user_cache_dir, "worktrees")

_git_object_name = re.compile("(^|/)([0-9a-fA-F]{4,}|(.+_)?HEAD)$")

_NOT_SPECIFIED = object()


class _TreeEntry(t.NamedTuple):
    mode: str
    type: str
    object: str
    file: str

    @classmethod
    def fromstring(cls, entry: str) -> _TreeEntry:
        info, file = entry.split("\t", maxsplit=1)
        mode, type, object = info.split(" ")
        return cls(mode, type, object, file)

    def tostring(self) -> str:
        return f"{self.mode} {self.type} {self.object}\t{self.file}"


class _WritableGitFile(t.BinaryIO):
    def __init__(
        self,
        tx: _GitTransaction,
        cachedir: pathlib.Path,
        path: pathlib.PurePosixPath,
    ) -> None:
        tx.record_pending_update(path, self)
        self.__tx = tx
        self.__path = path
        self.__file = cachedir.joinpath(path).open("wb")

    @property
    def mode(self) -> str:
        return self.__file.mode

    @property
    def name(self) -> str:
        return str(self.__path)

    def close(self) -> None:
        self.__file.close()
        self.__tx.record_update(self.__path)

    @property
    def closed(self) -> bool:
        return self.__file.closed

    def fileno(self) -> int:
        return self.__file.fileno()

    def flush(self) -> None:
        self.__file.flush()

    def isatty(self) -> bool:
        return self.__file.isatty()

    def read(self, n: int = -1) -> bytes:
        return self.__file.read(n)

    def readable(self) -> bool:
        return self.__file.readable()

    def readline(self, limit: int = -1) -> bytes:
        return self.__file.readline(limit)

    def readlines(self, hint: int = -1) -> list[bytes]:
        return self.__file.readlines(hint)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.__file.seek(offset, whence)

    def seekable(self) -> bool:
        return self.__file.seekable()

    def tell(self) -> int:
        return self.__file.tell()

    def truncate(self, size: int | None = None) -> int:
        return self.__file.truncate(size)

    def writable(self) -> bool:
        return self.__file.writable()

    def write(self, s: bytes) -> int:  # type: ignore[override] # ???
        return self.__file.write(s)

    def writelines(  # type: ignore[override] # ???
        self, lines: t.Iterable[bytes]
    ) -> None:
        self.__file.writelines(lines)

    def __enter__(self) -> _WritableGitFile:
        self.__file.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_trace):
        self.__file.__exit__(exc_type, exc_value, exc_trace)
        self.__tx.record_update(self.__path)

    def __iter__(self) -> _WritableGitFile:
        return self

    def __next__(self) -> bytes:
        return self.readline()


class _GitTransaction:
    __old_sha: str
    __unclosed_error = textwrap.dedent(
        """\
        %d file(s) is/are still opened.

        =======================================================================

        You still have open files when committing a transaction.

        Files that are still open CANNOT be committed and are LOST FOREVER!

        Fix your code to ``close()`` all opened files BEFORE control flows out
        of the ``with:`` block opened by ``write_transaction()``.

        Again: YOU HAVE JUST LOST DATA.

        =======================================================================
        """
    )

    def __init__(
        self,
        outer_transactor: cabc.Callable[..., t.ContextManager],
        filehandler: GitFileHandler,
        /,
        *,
        dry_run: bool,
        author_name: str | None,
        author_email: str | None,
        commit_msg: str,
        ignore_empty: bool,
        remote_branch: str | None,
        push: bool,
        push_options: cabc.Sequence[str],
        **kw: t.Any,
    ) -> None:
        self.__outer_context = outer_transactor(**kw)
        self.__handler = filehandler
        self.__dry_run = dry_run
        self.__commit_msg = commit_msg
        self.__ignore_empty = ignore_empty
        self.__push = push
        self.__push_opts = [f"--push-option={i}" for i in push_options]

        self.__gitenv: dict[str, str] = {}
        if author_name:
            self.__gitenv["GIT_AUTHOR_NAME"] = author_name
        if author_email:
            self.__gitenv["GIT_AUTHOR_EMAIL"] = author_email

        targetref = remote_branch or filehandler.revision

        if targetref == "HEAD":
            targetref = self.__resolve_head() or targetref

        if _git_object_name.search(targetref):
            raise ValueError(
                "Target ref looks like a git object, use a different"
                f" remote_branch: {targetref}"
            )
        if not targetref.startswith("refs/heads/"):
            targetref = "refs/heads/" + targetref

        self.__targetref = targetref
        self.__open_files: cabc.MutableMapping[
            tuple[int, pathlib.PurePosixPath], _WritableGitFile
        ] = weakref.WeakValueDictionary()

    def __enter__(self) -> cabc.Mapping[str, t.Any]:
        self.__old_sha = (
            self.__handler._git("rev-parse", self.__handler.revision)
            .decode("ascii")
            .strip()
        )
        if self.__handler._transaction is not None:
            raise RuntimeError("Another transaction is already open")
        self.__handler._transaction = self

        return self.__outer_context.__enter__()

    def __exit__(self, exc_type, exc_value, exc_trace):
        if exc_value is not None:
            self.__handler._transaction = None
            self.__handler._git("reset", "--hard", self.__old_sha)
            return self.__outer_context.__exit__(
                exc_type, exc_value, exc_trace
            )
        try:
            LOGGER.debug("Writing updated tree to database")
            tree = self.__write_tree()

            if self.__ignore_empty and tree == self.__get_old_tree_hash():
                LOGGER.debug("Not creating empty commit (ignore_empty=True)")
                return None

            LOGGER.debug("Creating commit object with tree %s", tree)
            commit = self.__commit(tree)
            if self.__dry_run:
                LOGGER.debug("Not updating branch pointers (dry_run=True)")
                return None

            LOGGER.debug("Updating ref %r to %s", self.__targetref, commit)
            self.__handler._git("reset", "--soft", commit)
            self.__update_target_ref(commit)

            if not self.__push:
                LOGGER.debug("Not pushing changes to remote (push=False)")
                return None

            LOGGER.debug("Pushing updated ref %r", self.__targetref)
            self.__push_updates("origin")
        finally:
            del self.__old_sha
            self.__handler._transaction = None
            self.__outer_context.__exit__(exc_type, exc_value, exc_trace)
        return None

    def record_update(self, filename: pathlib.PurePosixPath) -> None:
        """Record an updated file in the current transaction.

        Parameters
        ----------
        filename
            The path of the file, relative to the root of the repository.
        """
        assert self.__handler._transaction is self
        self.__handler._git("add", filename)

    def record_pending_update(
        self, filename: pathlib.PurePosixPath, file: _WritableGitFile
    ) -> None:
        self.__open_files[id(file), filename] = file

    def __commit(self, tree: str) -> str:
        """Commit ``tree`` as child commit of ``__target_ref``."""
        commit_hash = self.__handler._git(
            "commit-tree",
            tree,
            "-p",
            self.__old_sha,
            "-m",
            self.__commit_msg,
            env=self.__gitenv,
            encoding="ascii",
        )
        commit_hash = commit_hash.strip()

        LOGGER.debug("Created commit with hash %r", commit_hash)
        return commit_hash

    def __push_updates(self, remote: str) -> None:
        """Push the locally updated ``__target_ref`` to ``remote``."""
        self.__handler._git(
            "-c",
            f"remote.{remote}.mirror=false",
            "push",
            *self.__push_opts,
            "--",
            remote,
            self.__targetref,
        )

    def __resolve_head(self) -> str:
        """Resolve HEAD to a single symbolic name."""
        try:
            name = self.__handler._git(
                "rev-parse", "--symbolic-full-name", "HEAD", silent=True
            )
        except subprocess.CalledProcessError:
            name = b""
        return name.decode("utf-8").strip()

    def __update_target_ref(self, commit: str) -> None:
        """Update the local ``__target_ref``."""
        self.__handler._git(
            "update-ref",
            f"-m[{capellambse.__name__}] Commit by capellambse",
            self.__targetref,
            commit,
        )

    def __write_tree(self) -> str:
        """Write the tree built up in the git index to the database."""
        unclosed_files = 0
        for (_, filename), file in self.__open_files.items():
            if not file.closed:
                unclosed_files += 1
                LOGGER.warning("File is still open: %s", filename)
        if unclosed_files:
            LOGGER.critical(self.__unclosed_error, unclosed_files)

        tree_hash = self.__handler._git("write-tree", encoding="ascii").strip()
        LOGGER.debug("Created tree with hash %r", tree_hash)
        return tree_hash

    def __get_old_tree_hash(self) -> str:
        info = self.__handler._git("cat-file", "commit", self.__old_sha)
        for line in info.splitlines():
            if not line:
                break
            kw, hash = line.split(None, 1)
            if kw == b"tree":
                return hash.decode("ascii")

        raise AssertionError(f"No 'tree' in commit {self.__old_sha!r}")


class GitFileHandler(abc.FileHandler):
    """File handler for ``git://`` and related protocols.

    Parameters
    ----------
    revision
        The Git revision to check out. Either a branch or tag name, a
        full ref name, or the object name (i.e. hash) of a commit-ish.
    username
        The user name for authentication with the Git remote.
    password
        The password for authentication with the Git remote.
    identity_file
        Authenticate against the remote with the private key in this
        file. Defaults to using SSH's algorithm for determining a
        suitable key. (SSH only, ignored otherwise)
    known_hosts_file
        An OpenSSH-style ``known_hosts`` file containing the public key
        of the remote server. (SSH only, ignored otherwise)
    disable_cache
        Wipe the local cache and create a fresh, new clone.
    update_cache
        Contact the remote and make sure that the local cache is up to
        date. Note that setting this to ``False`` does not necessarily
        inhibit all attempts to contact the remote; it just disables the
        initial "fetch" operation. Later operations may still require to
        access the server, for example to download Git-LFS files.

    See Also
    --------
    capellambse.filehandler.abc.FileHandler :
        Documentation of common parameters.
    """

    username: str
    password: str
    identity_file: str
    known_hosts_file: str
    cache_dir: pathlib.Path
    """Path to the temporary work tree created by this file handler."""
    shallow: t.Final = False

    __fnz: object
    __has_lfs: bool
    __lfsfiles: dict[pathlib.PurePosixPath, bool]
    __repo: pathlib.Path
    __lockfile: pathlib.Path

    def __init__(
        self,
        path: str | os.PathLike,
        revision: str = "HEAD",
        username: str = "",
        password: str = "",
        identity_file: str = "",
        known_hosts_file: str = "",
        disable_cache: bool = False,
        update_cache: bool = True,
        *,
        subdir: str | pathlib.PurePosixPath = "/",
    ) -> None:
        super().__init__(path, subdir=subdir)
        self.disable_cache = disable_cache

        if bool(username) != bool(password):
            raise TypeError("Either specify username and password or neither")

        self.username = username
        self.password = password
        self.identity_file = identity_file
        self.known_hosts_file = known_hosts_file
        self.update_cache = update_cache

        self.cache_dir = None  # type: ignore[assignment]
        self.__hash, self.revision = self.__resolve_remote_ref(revision)

        self.__init_cache_dir()
        self.__init_worktree()

        self._transaction: _GitTransaction | None = None

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        path = capellambse.helpers.normalize_pure_path(
            filename, base=self.subdir
        )
        if "w" in mode:
            if self._transaction is None:
                raise abc.TransactionClosedError(
                    "Writing to git requires a transaction"
                )
            return _WritableGitFile(self._transaction, self.cache_dir, path)
        return open(self.cache_dir / path, "rb")

    def get_model_info(self) -> abc.HandlerInfo:
        def revparse(*args: str) -> str:
            return (
                self._git("rev-parse", *args, silent=True)
                .decode("utf-8", errors="surrogateescape")
                .strip()
            )

        return GitHandlerInfo(
            branch=revparse("--abbrev-ref", self.revision),
            url=str(self.path),
            revision=self.revision,
            rev_hash=self.__hash,
        )

    def write_transaction(
        self,
        dry_run: bool = False,
        author_name: str | None = None,
        author_email: str | None = None,
        commit_msg: str = "Changes made with python-capellambse",
        ignore_empty: bool = True,
        remote_branch: str | None = None,
        push: bool = True,
        push_options: cabc.Sequence[str] = (),
        **kw: t.Any,
    ) -> t.ContextManager[cabc.Mapping[str, t.Any]]:
        """Create a transaction that records all changes as a new commit.

        Parameters
        ----------
        author_name
            The name of the commit author.
        author_email
            The e-mail address of the commit author.
        commit_msg
            The commit message.
        dry_run
            If True, stop before updating the ``revision`` pointer. The
            commit will be created, but will not be part of any branch
            or tag.
        ignore_empty
            If True and the transaction did not actually change any
            files (i.e. the new commit would be tree-same with its
            parent), do not actually make a new commit.

            .. versionchanged:: 0.5.67
               Previous versions would create empty commits if no files
               changed. If you relied on that behavior (e.g. to trigger
               subsequent CI actions), use this option.
        remote_branch
            An alternative branch name to push to on the remote, instead
            of pushing back to the same branch. This is required if
            ``push`` is ``True`` and the ``revision`` that was passed to
            the constructor does not refer to a branch (or looks like a
            git object).

            Note: For convenience, ``refs/heads/`` will be prepended
            automatically to this name if it isn't already present. This
            also means that it is not possible to create tags or other
            types of refs; passing in something like ``refs/tags/v2.4``
            would result in the full ref name
            ``refs/heads/refs/tags/v2.4``.
        push
            Set to ``False`` to inhibit pushing the changes back.
        push_options
            Additional git push options. See ``--push-option`` in
            ``git-push(1)``. Ignored if ``push`` is ``False``.
        **kw
            Additional arguments are ignored.

        Raises
        ------
        ValueError
            - If a commit hash was used during loading, and no
              ``remote_branch`` was given
            - If the given ``remote_branch`` (or the final part of the
              originally given revision) looks like a git object
        """
        return _GitTransaction(
            super().write_transaction,
            self,
            dry_run=dry_run,
            author_name=author_name,
            author_email=author_email,
            commit_msg=commit_msg,
            ignore_empty=ignore_empty,
            remote_branch=remote_branch,
            push=push,
            push_options=push_options,
            **kw,
        )

    @property
    def rootdir(self) -> GitPath:
        """The root directory of the repository."""
        return GitPath(self, pathlib.PurePosixPath("."))

    def iterdir(
        self, path: str | pathlib.PurePosixPath = "."
    ) -> t.Iterator[GitPath]:
        """Iterate over the files in the given directory.

        Parameters
        ----------
        path
            The path to the directory to iterate over.
        """
        path = capellambse.helpers.normalize_pure_path(path, base=self.subdir)
        for subpath in self.cache_dir.joinpath(*path.parts).iterdir():
            yield GitPath(self, pathlib.PurePosixPath(path, subpath.name))

    @staticmethod
    def __cleanup_worktree(
        repo_root: pathlib.Path, worktree: pathlib.Path, /
    ) -> None:
        LOGGER.debug("Removing worktree at %s", worktree)
        subprocess.run(
            ["git", "worktree", "remove", "-f", str(worktree)],
            check=True,
            cwd=repo_root,
        )

    def __get_git_env(self) -> tuple[dict[str, str], list[str]]:
        git_env = os.environ.copy()
        git_cmd = ["git"]

        if not os.environ.get("GIT_ASKPASS"):
            path_to_askpass = (
                pathlib.Path(__file__).parent / "git_askpass.py"
            ).absolute()

            git_env["GIT_ASKPASS"] = str(path_to_askpass)

            try:
                os.chmod(path_to_askpass, 0o755)
            except OSError:
                LOGGER.info(
                    "Setting permission 755 for GIT_ASKPASS file failed"
                )

        if self.username and self.password:
            git_env["GIT_USERNAME"] = self.username
            git_env["GIT_PASSWORD"] = self.password

            git_cmd += ["-c", "credential.helper="]

        if self.identity_file and self.known_hosts_file:
            ssh_command = [
                "ssh",
                "-i",
                self.identity_file,
                f"-oUserKnownHostsFile={self.known_hosts_file}",
            ]
            git_env["GIT_SSH_COMMAND"] = shlex.join(ssh_command)

            git_cmd += ["-c", "credential.helper="]

        return git_env, git_cmd

    def __is_local_repo(self) -> bool:
        return (
            isinstance(self.path, pathlib.Path)
            or str(self.path).startswith("file://")
            or pathlib.Path(self.path).is_absolute()
        )

    def __init_cache_dir(self) -> None:
        if self.__is_local_repo():
            self.__init_cache_dir_local()
        else:
            self.__init_cache_dir_remote()

    def __init_cache_dir_local(self) -> None:
        if isinstance(self.path, pathlib.Path):
            path = self.path
        elif str(self.path).startswith("file://"):
            urlpath = urllib.parse.urlparse(str(self.path)).path
            urlpath = urllib.parse.unquote(urlpath)
            windows = isinstance(pathlib.Path(), pathlib.WindowsPath)
            path = pathlib.Path(urlpath[1:] if windows else urlpath)
        else:
            path = pathlib.Path(self.path)
        self.__repo = self.cache_dir = path.resolve()
        gitdir = self.__git_nolock("rev-parse", "--git-dir", encoding="utf-8")
        assert isinstance(gitdir, str)
        self.__lockfile = (
            pathlib.Path(self.cache_dir, gitdir.strip())
            .resolve()
            .joinpath("capellambse.lock")
        )

    def __init_cache_dir_remote(self) -> None:
        slug_pattern = '[\x00-\x1f\x7f"*/:<>?\\|]+'
        path_slug = re.sub(slug_pattern, "-", str(self.path))
        hashpath = str(self.path).encode("utf-8", errors="surrogatepass")

        path_hash = hashlib.sha256(hashpath, usedforsecurity=False).hexdigest()
        old_dir = CACHEBASE.joinpath(path_hash, path_slug)

        path_hash = hashlib.blake2s(
            hashpath, digest_size=12, usedforsecurity=False
        ).hexdigest()
        self.__repo = self.cache_dir = CACHEBASE.joinpath(path_hash, path_slug)
        self.__lockfile = self.__repo.joinpath("capellambse.lock")

        if old_dir.exists():
            LOGGER.debug("Moving cache from %s to %s", old_dir, self.cache_dir)
            self.cache_dir.parent.mkdir(parents=True, exist_ok=True)
            os.rename(old_dir, self.cache_dir)
            try:
                old_dir.parent.rmdir()
            except OSError as err:
                if err.errno != errno.ENOTEMPTY:
                    raise

        if self.cache_dir.exists() and self.disable_cache:
            shutil.rmtree(str(self.cache_dir))

        self.cache_dir.mkdir(parents=True, exist_ok=True)
        with capellambse.helpers.flock(self.__repo / "capellambse.lock"):
            update_cache = self.update_cache
            if not self.cache_dir.joinpath("config").exists():
                LOGGER.debug("Creating a new git repo in %s", self.cache_dir)
                self.__git_nolock(
                    "-c", "init.defaultBranch=master", "init", "--bare"
                )
                self.__git_nolock(
                    "remote", "add", "--mirror=fetch", "origin", self.path
                )
                update_cache = True

            if update_cache:
                LOGGER.debug("Updating cache at %s", self.cache_dir)
                fetchopts = ["--filter=tree:0"]
                if (self.cache_dir / "shallow").exists():
                    fetchopts.append("--unshallow")
                fetchspec = f"+{self.revision}"
                if not _git_object_name.search(self.revision):
                    fetchspec += f":{self.revision}"
                self.__git_nolock("fetch", *fetchopts, self.path, fetchspec)

            self.__git_nolock("config", "extensions.worktreeConfig", "true")

    def __resolve_default_branch(self) -> tuple[str, str]:
        """Resolve the default branch name and its hash on the remote."""
        LOGGER.debug("Resolving default branch on remote %s", self.path)
        listing = self.__git_nolock(
            "ls-remote", "--symref", self.path, "HEAD", encoding="utf-8"
        )
        assert isinstance(listing, str)
        match = re.search(
            r"ref: (refs/heads/.*)\s+HEAD\s+([0-9a-fA-F]+)", listing
        )
        if match:
            refname = match.group(1)
            hash = match.group(2)
            return hash, refname
        raise ValueError("Failed to resolve default branch on remote")

    def __resolve_remote_ref(self, ref: str) -> tuple[str, str]:
        """Resolve the given ``ref`` on the remote."""
        if ref == "HEAD":
            try:
                return self.__resolve_default_branch()
            except ValueError:
                pass
        LOGGER.debug("Resolving ref %r on remote %s", ref, self.path)
        listing = self.__git_nolock(
            "ls-remote", self.path, ref, encoding="utf-8"
        )
        assert isinstance(listing, str)
        if not listing:
            if not _git_object_name.search(ref):
                raise ValueError(f"Ref does not exist on remote: {ref}")
            LOGGER.debug("Ref %r not found, assuming object name", ref)
            return ref, ref
        refs = [i.split("\t") for i in listing.strip().split("\n")]
        if len(refs) > 1:
            exact = [i for i in refs if i[1] == ref]
            if len(exact) != 1:
                raise ValueError(
                    f"Ambiguous ref name {ref}, found {len(refs)}:"
                    f" {', '.join(i[1] for i in refs)}"
                )
            refs = exact

        hash, refname = refs[0]
        LOGGER.debug(
            "Resolved ref %r as remote ref %r (%s)", ref, refname, hash
        )
        return hash, refname

    def __init_worktree(self) -> None:
        WTBASE.mkdir(0o700, parents=True, exist_ok=True)
        worktree = pathlib.Path(
            tempfile.mkdtemp(None, f"capellambse-{os.getpid()}-", WTBASE)
        )
        LOGGER.debug("Setting up a worktree at %s", worktree)
        try:
            self._git(
                "worktree",
                "add",
                "--detach",
                worktree,
                self.__hash,
            )
        except:
            os.rmdir(worktree)
            raise
        self.cache_dir = worktree
        self.__fnz = weakref.finalize(
            self, self.__cleanup_worktree, self.__repo, worktree
        )

        gitdir = self.__git_nolock("rev-parse", "--git-dir", encoding="utf-8")
        assert isinstance(gitdir, str)
        self.__repo = pathlib.Path(self.cache_dir, gitdir.strip()).resolve()

        if not self.__is_local_repo():
            try:
                self.__git_nolock("lfs", "env", silent=True)
            except subprocess.CalledProcessError:
                LOGGER.debug("LFS not installed, skipping setup")
            else:
                LOGGER.debug("LFS support detected, registering filter")
                self.__git_nolock(
                    "config",
                    "--worktree",
                    "core.hooksPath",
                    self.__repo.joinpath("hooks"),
                )
                self.__git_nolock("lfs", "install", "--worktree", "--force")
                self.__git_nolock("config", "--worktree", "core.bare", "false")
                self._git("lfs", "pull")

    @t.overload
    def _git(
        self,
        *cmd: t.Any,
        encoding: str,
        env: cabc.Mapping[str, str] | None = ...,
        silent: bool = ...,
        **kw: t.Any,
    ) -> str: ...
    @t.overload
    def _git(
        self,
        *cmd: t.Any,
        encoding: None = ...,
        env: cabc.Mapping[str, str] | None = ...,
        silent: bool = ...,
        **kw: t.Any,
    ) -> bytes: ...
    def _git(
        self,
        *cmd: t.Any,
        env: cabc.Mapping[str, str] | None = None,
        silent: bool = False,
        **kw: t.Any,
    ) -> bytes | str:
        with capellambse.helpers.flock(self.__lockfile):
            return self.__git_nolock(*cmd, env=env, silent=silent, **kw)

    def __git_nolock(
        self,
        *cmd: t.Any,
        env: cabc.Mapping[str, str] | None = None,
        silent: bool = False,
        **kw: t.Any,
    ) -> bytes | str:
        LOGGER.debug("Running command %s", cmd)
        returncode = 0
        stderr = None

        git_env, git_cmd = self.__get_git_env()

        try:
            proc = subprocess.run(
                git_cmd + [str(i) for i in cmd],
                capture_output=True,
                check=True,
                cwd=self.cache_dir,
                env={**git_env, **(env or {})},
                **kw,
            )
            returncode = proc.returncode
            stderr = proc.stderr
            return proc.stdout
        except subprocess.CalledProcessError as err:
            returncode = err.returncode
            stderr = err.stderr
            raise
        finally:
            if returncode != 0 and not silent:
                level = logging.ERROR
            else:
                level = logging.DEBUG

            if stderr:
                if isinstance(stderr, bytes):
                    stderr = stderr.decode("utf-8")

                for line in stderr.splitlines():
                    LOGGER.getChild("git").log(level, "%s", line)
            LOGGER.log(level, "Exit status: %d", returncode)


class GitPath(abc.FilePath[GitFileHandler]):
    def is_dir(self) -> bool:
        return self._parent.cache_dir.joinpath(self._path).is_dir()

    def is_file(self) -> bool:
        return self._parent.cache_dir.joinpath(self._path).is_file()


@dataclasses.dataclass
class GitHandlerInfo(abc.HandlerInfo):
    branch: str | None
    revision: str
    rev_hash: str



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Metamodel definitions for Capella models."""

from . import activity as activity
from . import behavior as behavior
from . import capellacommon as capellacommon
from . import capellacore as capellacore
from . import capellamodeller as capellamodeller
from . import cs as cs
from . import epbs as epbs
from . import fa as fa
from . import information as information
from . import interaction as interaction
from . import la as la
from . import libraries as libraries
from . import modellingcore as modellingcore
from . import oa as oa
from . import pa as pa
from . import sa as sa



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Operational Analysis layer."""

from __future__ import annotations

import sys
import typing as t
import warnings

import capellambse.model as m

from . import (
    activity,
    capellacommon,
    capellacore,
    cs,
    fa,
    information,
    interaction,
    modellingcore,
)
from . import namespaces as ns

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

NS = ns.OA


class OperationalAnalysis(cs.BlockArchitecture):
    """Provides access to the OperationalAnalysis layer of the model."""

    role_pkg = m.Containment["RolePkg"]("ownedRolePkg", (NS, "RolePkg"))
    entity_pkg = m.Single["EntityPkg"](
        m.Containment("ownedEntityPkg", (NS, "EntityPkg"))
    )
    concept_pkg = m.Containment["ConceptPkg"](
        "ownedConceptPkg", (NS, "ConceptPkg")
    )

    activity_pkg = m.Alias["OperationalActivityPkg"]("function_pkg")

    @property
    def all_activities(self) -> m.ElementList[OperationalActivity]:
        return self._model.search((NS, "OperationalActivity"), below=self)

    @property
    def all_processes(self) -> m.ElementList[OperationalProcess]:
        return self._model.search((NS, "OperationalProcess"), below=self)

    @property
    def all_actors(self) -> m.ElementList[Entity]:
        return self._model.search((NS, "Entity")).by_is_actor(True)

    @property
    def all_entities(self) -> m.ElementList[Entity]:
        return self._model.search((NS, "Entity"), below=self)

    @property
    def all_activity_exchanges(self) -> m.ElementList[fa.FunctionalExchange]:
        return self._model.search((ns.FA, "FunctionalExchange"), below=self)

    @property
    def all_entity_exchanges(self) -> m.ElementList[CommunicationMean]:
        return self._model.search((NS, "CommunicationMean"), below=self)

    @property
    def all_operational_processes(self) -> m.ElementList[OperationalProcess]:
        return self._model.search(OperationalProcess, below=self)

    @property
    @deprecated(
        (
            "OperationalActivity.root_activity can only handle a single"
            " OperationalActivity, use .activity_pkg.activities directly instead"
        ),
        category=FutureWarning,
    )
    def root_activity(self) -> OperationalActivity:
        pkg = self.activity_pkg
        if pkg is None:
            raise m.BrokenModelError(
                "OperationalAnalysis has no root ActivityPkg"
            )
        assert isinstance(pkg, OperationalActivityPkg)
        candidates = pkg.activities
        if len(candidates) < 1:
            raise m.BrokenModelError(
                "ActivityPkg does not contain any Activities"
            )
        if len(candidates) > 1:
            raise RuntimeError(
                "Expected 1 object for OperationalAnalysis.root_activity,"
                f" got {len(candidates)}"
            )
        return candidates[0]

    @property
    @deprecated(
        (
            "OperationalActivity.root_entity can only handle a single"
            " Entity, use .entity_pkg.entities directly instead"
        ),
        category=FutureWarning,
    )
    def root_entity(self) -> Entity:
        pkg = self.entity_pkg
        if pkg is None:
            raise m.BrokenModelError(
                "OperationalAnalysis has no root EntityPkg"
            )
        candidates = pkg.entities
        if len(candidates) < 1:
            raise m.BrokenModelError("Root EntityPkg is empty")
        if len(candidates) > 1:
            raise RuntimeError(
                "Expected 1 object for OperationalAnalysis.root_entity,"
                f" got {len(candidates)}"
            )
        return candidates[0]

    diagrams = m.DiagramAccessor(
        "Operational Analysis", cacheattr="_MelodyModel__diagram_cache"
    )

    if not t.TYPE_CHECKING:
        entity_package = m.DeprecatedAccessor("entity_pkg")
        activity_package = m.DeprecatedAccessor("activity_pkg")
        capability_package = m.DeprecatedAccessor("capability_pkg")


class OperationalScenario(capellacore.NamedElement, abstract=True):
    context = m.StringPOD("context")
    objective = m.StringPOD("objective")


class OperationalActivityPkg(fa.FunctionPkg):
    _xmltag = "ownedFunctionPkg"

    activities = m.Containment["OperationalActivity"](
        "ownedOperationalActivities", (NS, "OperationalActivity")
    )
    packages = m.Containment["OperationalActivityPkg"](
        "ownedOperationalActivityPkgs", (NS, "OperationalActivityPkg")
    )
    owner = m.Single["Entity"](m.Backref((NS, "Entity"), "activities"))


class OperationalActivity(fa.AbstractFunction):
    _xmltag = "ownedOperationalActivities"

    packages = m.Containment["OperationalActivityPkg"](
        "ownedOperationalActivityPkgs", (NS, "OperationalActivityPkg")
    )
    activities = m.Alias["m.ElementList[OperationalActivity]"]("functions")
    inputs = m.Backref["fa.FunctionalExchange"](  # type: ignore[assignment]
        (ns.FA, "FunctionalExchange"), "target"
    )
    outputs = m.Backref["fa.FunctionalExchange"](  # type: ignore[assignment]
        (ns.FA, "FunctionalExchange"), "source"
    )
    realizing_system_functions = m.Backref["sa.SystemFunction"](
        (ns.SA, "SystemFunction"), "realized_operational_activities"
    )

    owner = m.Single["Entity"](m.Backref((NS, "Entity"), "activities"))

    related_exchanges = m.Backref[fa.FunctionalExchange](
        (ns.FA, "FunctionalExchange"), "source", "target"
    )


class OperationalProcess(fa.FunctionalChain):
    pass


class Swimlane(capellacore.NamedElement, activity.ActivityPartition):
    pass


class OperationalCapabilityPkg(capellacommon.AbstractCapabilityPkg):
    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = m.Containment["OperationalCapability"](
        "ownedOperationalCapabilities", (NS, "OperationalCapability")
    )
    packages = m.Containment["OperationalCapabilityPkg"](
        "ownedOperationalCapabilityPkgs", (NS, "OperationalCapabilityPkg")
    )
    capability_configurations = m.Containment["CapabilityConfiguration"](
        "ownedCapabilityConfigurations", (NS, "CapabilityConfiguration")
    )
    concept_compliances = m.Containment["ConceptCompliance"](
        "ownedConceptCompliances", (NS, "ConceptCompliance")
    )
    complies_with_concepts = m.Allocation["Concept"](
        "ownedConceptCompliances",
        (NS, "ConceptCompliance"),
        (NS, "Concept"),
        attr="complyWithConcept",
        backattr="compliantCapability",
    )


class OperationalCapability(
    interaction.AbstractCapability, capellacore.Namespace
):
    """A capability in the OperationalAnalysis layer."""

    _xmltag = "ownedOperationalCapabilities"

    compliances = m.Association["ConceptCompliance"](
        (NS, "ConceptCompliance"), "compliances"
    )
    configurations = m.Association["CapabilityConfiguration"](
        (NS, "CapabilityConfiguration"), "configurations"
    )
    entity_involvements = m.Containment[
        "EntityOperationalCapabilityInvolvement"
    ](
        "ownedEntityOperationalCapabilityInvolvements",
        (NS, "EntityOperationalCapabilityInvolvement"),
    )
    involved_entities = m.Allocation["Entity"](
        "ownedEntityOperationalCapabilityInvolvements",
        (NS, "EntityOperationalCapabilityInvolvement"),
        (NS, "Entity"),
        attr="involved",
        legacy_by_type=True,
    )
    involved_activities = m.Alias["m.ElementList[OperationalActivity]"](
        "involved_functions"
    )
    involved_processes = m.Alias["m.ElementList[OperationalProcess]"](
        "involved_chains"
    )
    owned_processes = m.Alias["m.ElementList[OperationalProcess]"](
        "functional_chains"
    )


class ActivityAllocation(capellacore.Allocation):
    pass


class RolePkg(capellacore.Structure):
    packages = m.Containment["RolePkg"]("ownedRolePkgs", (NS, "RolePkg"))
    roles = m.Containment["Role"]("ownedRoles", (NS, "Role"))


class Role(information.AbstractInstance):
    assembly_usages = m.Containment["RoleAssemblyUsage"](
        "ownedRoleAssemblyUsages", (NS, "RoleAssemblyUsage")
    )
    activity_allocations = m.Containment["ActivityAllocation"](
        "ownedActivityAllocations", (NS, "ActivityAllocation")
    )


class RoleAssemblyUsage(capellacore.NamedElement):
    child = m.Association["Role"]((NS, "Role"), "child")


class RoleAllocation(capellacore.Allocation):
    pass


class EntityPkg(cs.ComponentPkg):
    _xmltag = "ownedEntityPkg"

    entities = m.Containment["Entity"]("ownedEntities", (NS, "Entity"))
    packages = m.Containment["EntityPkg"]("ownedEntityPkgs", (NS, "EntityPkg"))
    locations = m.Containment["Location"]("ownedLocations", (NS, "Location"))
    communication_means = m.Alias["m.ElementList[CommunicationMean]"](
        "exchanges"
    )


class AbstractConceptItem(cs.Component, abstract=True):
    composing_links = m.Association["ItemInConcept"](
        (NS, "ItemInConcept"), "composingLinks"
    )


class Entity(
    AbstractConceptItem,
    modellingcore.InformationsExchanger,
    capellacore.InvolvedElement,
):
    """An Entity in the OperationalAnalysis layer."""

    _xmltag = "ownedEntities"

    organisational_unit_memberships = m.Association[
        "OrganisationalUnitComposition"
    ]((NS, "OrganisationalUnitComposition"), "organisationalUnitMemberships")
    actual_location = m.Association["Location"](
        (NS, "Location"), "actualLocation"
    )
    entities = m.Containment["Entity"]("ownedEntities", (NS, "Entity"))
    communication_means = m.Containment["CommunicationMean"](
        "ownedCommunicationMeans", (NS, "CommunicationMean")
    )
    exchanges = m.Alias["m.ElementList[CommunicationMean]"](
        "communication_means"
    )
    activities = m.Allocation["OperationalActivity"](
        "ownedFunctionalAllocation",
        (ns.FA, "ComponentFunctionalAllocation"),
        (NS, "OperationalActivity"),
        attr="targetElement",
        backattr="sourceElement",
    )
    capabilities = m.Backref["OperationalCapability"](
        (NS, "OperationalCapability"), "involved_entities"
    )
    related_exchanges = m.Backref["CommunicationMean"](
        (NS, "CommunicationMean"), "source", "target"
    )
    realizing_system_components = m.Backref["sa.SystemComponent"](
        (ns.SA, "SystemComponent"), "realized_operational_entities"
    )

    @property
    def inputs(self) -> m.ElementList[CommunicationMean]:
        return self._model.search((NS, "CommunicationMean")).by_target(self)

    @property
    def outputs(self) -> m.ElementList[CommunicationMean]:
        return self._model.search((NS, "CommunicationMean")).by_source(self)


class ConceptPkg(capellacore.Structure):
    packages = m.Containment["ConceptPkg"](
        "ownedConceptPkgs", (NS, "ConceptPkg")
    )
    concepts = m.Containment["Concept"]("ownedConcepts", (NS, "Concept"))


class Concept(capellacore.NamedElement):
    compliances = m.Association["ConceptCompliance"](
        (NS, "ConceptCompliance"), "compliances"
    )
    composite_links = m.Containment["ItemInConcept"](
        "compositeLinks", (NS, "ItemInConcept")
    )


class ConceptCompliance(capellacore.Relationship):
    comply_with_concept = m.Single["Concept"](
        m.Association((NS, "Concept"), "complyWithConcept")
    )
    compliant_capability = m.Single["OperationalCapability"](
        m.Association((NS, "OperationalCapability"), "compliantCapability")
    )


class ItemInConcept(capellacore.NamedElement):
    concept = m.Single["Concept"](m.Association((NS, "Concept"), "concept"))
    item = m.Single["AbstractConceptItem"](
        m.Association((NS, "AbstractConceptItem"), "item")
    )


class CommunityOfInterest(capellacore.NamedElement):
    community_of_interest_compositions = m.Containment[
        "CommunityOfInterestComposition"
    ](
        "communityOfInterestCompositions",
        (NS, "CommunityOfInterestComposition"),
    )


class CommunityOfInterestComposition(capellacore.NamedElement):
    community_of_interest = m.Association["CommunityOfInterest"](
        (NS, "CommunityOfInterest"), "communityOfInterest"
    )
    interested_organisational_unit = m.Association["OrganisationalUnit"](
        (NS, "OrganisationalUnit"), "interestedOrganisationUnit"
    )


class OrganisationalUnit(capellacore.NamedElement):
    organisational_unit_compositions = m.Containment[
        "OrganisationalUnitComposition"
    ]("organisationalUnitCompositions", (NS, "OrganisationalUnitComposition"))
    community_of_interest_memberships = m.Association[
        "CommunityOfInterestComposition"
    ]((NS, "CommunityOfInterestComposition"), "communityOfInterestMemberships")


class OrganisationalUnitComposition(capellacore.NamedElement):
    organisational_unit = m.Association["OrganisationalUnit"](
        (NS, "OrganisationalUnit"), "organisationalUnit"
    )
    participating_entity = m.Association["Entity"](
        (NS, "Entity"), "participatingEntity"
    )


class Location(AbstractConceptItem):
    location_description = m.StringPOD("locationDescription")
    located_entities = m.Association["Entity"](
        (NS, "Entity"), "locatedEntities"
    )


class CapabilityConfiguration(AbstractConceptItem):
    configured_capability = m.Association["OperationalCapability"](
        (NS, "OperationalCapability"), "configuredCapability"
    )


# NOTE: CommunicationMean should directly inherit from NamedRelationship,
# however this would result in an MRO conflict that cannot be resolved.
# Therefore we only inherit from ComponentExchange, copy the only missing
# definition (naming_rules), and register it as virtual subclass.
class CommunicationMean(fa.ComponentExchange):
    """An operational entity exchange."""

    _xmltag = "ownedComponentExchanges"

    # Taken from NamedRelationship, see note above
    naming_rules = m.Containment["capellacore.NamingRule"](
        "namingRules", (ns.CAPELLACORE, "NamingRule")
    )

    allocated_interactions = m.Alias["m.ElementList[fa.FunctionalExchange]"](
        "allocated_functional_exchanges"
    )


capellacore.NamedRelationship.register(CommunicationMean)


class EntityOperationalCapabilityInvolvement(capellacore.Involvement):
    pass


if not t.TYPE_CHECKING:

    def __getattr__(name):
        if name == "AbstractEntity":
            warnings.warn(
                "AbstractEntity has been merged into Entity",
                DeprecationWarning,
                stacklevel=2,
            )
            return Entity
        raise AttributeError(name)


from . import sa  # noqa: F401
